import json

import plotly.graph_objects as go

import nextmv


def create_visuals(name: str, radius: float, distance: float) -> list[nextmv.Asset]:
    """Create a Plotly bar chart with radius and distance for a planet."""

    fig = go.Figure()
    fig.add_trace(
        go.Bar(x=[name], y=[radius], name="Radius (km)", marker_color="red", opacity=0.5),
    )
    fig.add_trace(
        go.Bar(x=[name], y=[distance], name="Distance (Millions km)", marker_color="blue", opacity=0.5),
    )
    fig.update_layout(
        title="Radius and Distance by Planet", xaxis_title="Planet", yaxis_title="Values", barmode="group"
    )
    fig = fig.to_json()

    assets = [
        nextmv.Asset(
            name="Plotly example",
            content_type="json",
            visual=nextmv.Visual(
                visual_schema=nextmv.VisualSchema.PLOTLY,
                visual_type="custom-tab",
                label="Charts",
            ),
            content=[json.loads(fig)],
        )
    ]

    return assets



"""
This module defines the configuration delete command for the Nextmv CLI.
"""

from typing import Annotated

import rich
import typer
from rich.prompt import Confirm

from nextmv.cli.configuration.config import load_config, save_config
from nextmv.cli.error import error

# Set up subcommand application.
app = typer.Typer()


@app.command()
def delete(
    profile: Annotated[  # Similar to nextmv.cli.options.ProfileOption but with different help text.
        str,
        typer.Option(
            "--profile",
            "-p",
            help="Profile name to delete.",
            envvar="NEXTMV_PROFILE",
            metavar="PROFILE_NAME",
        ),
    ],
) -> None:
    """
    Delete a profile from the configuration.

    [bold][underline]Examples[/underline][/bold]

    - Delete a profile named [magenta]hare[/magenta].
        $ [green]nextmv configuration delete --profile hare[/green]
    """
    config = load_config()
    if profile not in config:
        error(f"Profile [bold magenta]{profile}[/bold magenta] does not exist.")

    confirm = Confirm.ask(
        f"Are you sure you want to delete profile [bold magenta]{profile}[/bold magenta]? This action cannot be undone",
        default=False,
    )

    if not confirm:
        rich.print(f":bulb: Profile [bold magenta]{profile}[/bold magenta] will not be deleted.")
        return

    del config[profile]
    save_config(config)

    rich.print(f":white_check_mark: Profile [bold magenta]{profile}[/bold magenta] deleted successfully.")



"""
Logger module that writes to stderr.

This module provides utilities for redirecting standard output to standard error
and for writing log messages directly to stderr.

Functions
---------
redirect_stdout
    Redirect all messages written to stdout to stderr.
reset_stdout
    Reset stdout to its original value.
log
    Log a message to stderr.
"""

import sys

# Original stdout reference held when redirection is active
__original_stdout = None
# Flag to track if stdout has been redirected
__stdout_redirected = False


def redirect_stdout() -> None:
    """
    Redirect all messages written to stdout to stderr.

    You can import the `redirect_stdout` function directly from `nextmv`:

    ```python
    from nextmv import redirect_stdout
    ```

    This function captures the current sys.stdout and replaces it with sys.stderr.
    When redirection is no longer needed, call `reset_stdout()` to restore the
    original stdout.

    Examples
    --------
    >>> redirect_stdout()
    >>> print("This will go to stderr")
    >>> reset_stdout()
    >>> print("This will go to stdout")
    """

    global __original_stdout, __stdout_redirected
    if __stdout_redirected:
        return
    __stdout_redirected = True

    __original_stdout = sys.stdout
    sys.stdout = sys.stderr


def reset_stdout() -> None:
    """
    Reset stdout to its original value.

    You can import the `reset_stdout` function directly from `nextmv`:

    ```python
    from nextmv import reset_stdout
    ```

    This function should always be called after `redirect_stdout()` to avoid
    unexpected behavior. It restores the original stdout that was captured
    during redirection.

    Examples
    --------
    >>> redirect_stdout()
    >>> print("This will go to stderr")
    >>> reset_stdout()
    >>> print("This will go to stdout")
    """
    global __original_stdout, __stdout_redirected
    if not __stdout_redirected:
        return
    __stdout_redirected = False

    if __original_stdout is None:
        sys.stdout = sys.__stdout__
        return

    sys.stdout = __original_stdout
    __original_stdout = None


def log(message: str) -> None:
    """
    Log a message to stderr.

    You can import the `log` function directly from `nextmv`:

    ```python
    from nextmv import log
    ```

    Parameters
    ----------
    message : str
        The message to log.

    Examples
    --------
    >>> log("An error occurred")
    An error occurred
    """

    print(message, file=sys.stderr)



import unittest

from nextmv.base_model import BaseModel


class Foo(BaseModel):
    bar: str
    baz: int | None = None


class Roh(BaseModel):
    foo: Foo
    qux: list[str] | None = None
    lorem: str | None = None


class TestBaseModel(unittest.TestCase):
    valid_dict = {
        "foo": {
            "bar": "bar",
            "baz": 1,
        },
        "qux": ["qux"],
        "lorem": "lorem",
    }

    def test_from_dict(self):
        roh = Roh.from_dict(self.valid_dict)
        self.assertTrue(isinstance(roh, Roh))
        self.assertTrue(isinstance(roh.foo, Foo))

    def test_change_attributes(self):
        roh = Roh.from_dict(self.valid_dict)
        self.assertEqual(roh.foo.bar, "bar")

        different_str = "different_str"
        roh.foo.bar = different_str
        self.assertEqual(roh.foo.bar, different_str)

    def test_invalid_dict(self):
        invalid = {
            "foo": {
                "bar": "bar",
                "baz": "1",
            },
            "qux": "qux",
            "lorem": "lorem",
        }
        with self.assertRaises(ValueError):
            Roh.from_dict(invalid)

    def test_to_dict(self):
        some_none = {
            "foo": {
                "bar": "bar",
            },
            "lorem": "lorem",
        }
        roh = Roh.from_dict(some_none)
        parsed = roh.to_dict()
        self.assertEqual(parsed, some_none)



"""
Application module for interacting with local Nextmv applications.

This module provides functionality to interact with applications in Nextmv,
including application management, running applications, and managing inputs.

Classes
-------
Application
    Class for interacting with local Nextmv applications.
"""

import json
import os
import shutil
import tempfile
import webbrowser
from dataclasses import dataclass
from datetime import datetime, timezone
from typing import Any

from nextmv import cloud
from nextmv._serialization import deflated_serialize_json
from nextmv.base_model import BaseModel
from nextmv.input import INPUTS_KEY, Input, InputFormat
from nextmv.local.local import (
    DEFAULT_INPUT_JSON_FILE,
    DEFAULT_OUTPUT_JSON_FILE,
    LOGS_FILE,
    LOGS_KEY,
    NEXTMV_DIR,
    RUNS_KEY,
)
from nextmv.local.runner import run
from nextmv.logger import log
from nextmv.manifest import Manifest, default_python_manifest
from nextmv.options import Options
from nextmv.output import ASSETS_KEY, OUTPUTS_KEY, SOLUTIONS_KEY, STATISTICS_KEY, OutputFormat
from nextmv.polling import DEFAULT_POLLING_OPTIONS, PollingOptions, poll
from nextmv.run import (
    ErrorLog,
    Format,
    FormatInput,
    Run,
    RunConfiguration,
    RunInformation,
    RunResult,
    SyncedRun,
    TrackedRun,
    TrackedRunStatus,
)
from nextmv.safe import safe_id
from nextmv.status import StatusV2


@dataclass
class Application:
    """
    A decision model that can be executed.

    You can import the `Application` class directly from `local`:

    ```python
    from nextmv.local import Application
    ```

    This class represents an application in Nextmv, providing methods to
    interact with the application, run it with different inputs, manage
    versions, instances, experiments, and more.

    Parameters
    ----------
    src : str
        Source of the application, when initialized locally. An application's
        source typically refers to the directory containing the `app.yaml`
        manifest.
    description : Optional[str], default=None
        Description of the application.

    Examples
    --------
    >>> from nextmv.local import Application
    >>> app = Application(src="path/to/app")
    >>> # Retrieve an app's run result
    >>> result = app.run_result("run-id")
    """

    src: str
    """
    Source of the application, when initialized locally. An application's
    source typically refers to the directory containing the `app.yaml`
    manifest.
    """

    description: str | None = None
    """Description of the application."""
    manifest: Manifest | None = None
    """
    Manifest of the application. A manifest is a file named `app.yaml` that
    must be present at the root of the application's `src` directory. If the
    app is initialized, and a manifest is not present, a default Python
    manifest will be created, using the `nextmv.default_python_manifest`
    function. If you specify this argument, and a manifest file is already
    present in the `src` directory, the provided manifest will override the
    existing one.
    """

    def __post_init__(self):
        """
        Validate the presence of the manifest in the application.
        """

        if self.manifest is not None:
            self.manifest.to_yaml(self.src)

            return

        try:
            manifest = Manifest.from_yaml(self.src)
            self.manifest = manifest

            return

        except Exception:
            manifest = default_python_manifest()
            self.manifest = manifest
            manifest.to_yaml(self.src)

            return

    @classmethod
    def initialize(
        cls,
        src: str | None = None,
        description: str | None = None,
        destination: str | None = None,
    ) -> "Application":
        """
        Initialize a sample Nextmv application, locally.

        This method will create a new application in the local file system. The
        application is a dir with the name given by `src` (it becomes the
        _source_ of the app), under the location given by `destination`. If the
        `destination` parameter is not specified, the current working directory
        is used as default. This method will scaffold the application with the
        necessary files and directories to have an opinionated structure for
        your decision model. Once the application is initialized, you are
        encouraged to complete it with the decision model itself, so that the
        application can be run locally.

        If the `src` parameter is not provided, a random name will be generated
        for the application.

        Parameters
        ----------
        src : str, optional
            Source (ID, name) of the application. Will be generated if not
            provided.
        description : str, optional
            Description of the application.
        destination : str, optional
            Destination directory where the application will be initialized. If
            not provided, the current working directory will be used.

        Returns
        -------
        Application
            The initialized application instance.
        """

        destination_dir = os.getcwd() if destination is None else destination
        app_id = src if src is not None else safe_id("app")

        # Create the new directory with the given name.
        app_src = os.path.join(destination_dir, app_id)
        if os.path.exists(app_src):
            raise FileExistsError(f"destination dir for src already exists: {app_src}")

        os.makedirs(app_src, exist_ok=False)

        # Get the path to the initial app structure template.
        current_file_dir = os.path.dirname(os.path.abspath(__file__))
        initial_app_structure_path = os.path.join(current_file_dir, "..", "default_app")
        initial_app_structure_path = os.path.normpath(initial_app_structure_path)

        # Copy everything from initial_app_structure to the new directory.
        if os.path.exists(initial_app_structure_path):
            shutil.copytree(initial_app_structure_path, app_src, dirs_exist_ok=True)

        return cls(
            src=app_src,
            description=description,
        )

    def list_runs(self) -> list[Run]:
        """
        List all runs for the application.

        Returns
        -------
        list[Run]
            A list of all runs associated with the application.
        """

        runs_dir = os.path.join(self.src, NEXTMV_DIR, RUNS_KEY)
        if not os.path.exists(runs_dir):
            raise ValueError(f"`.nextmv/runs` dir does not exist at app source: {self.src}")

        dirs = os.listdir(runs_dir)
        if not dirs:
            return []

        run_ids = [d for d in dirs if os.path.isdir(os.path.join(runs_dir, d))]
        if not run_ids:
            return []

        runs = []
        for run_id in run_ids:
            info = self.run_metadata(run_id=run_id)
            run = info.to_run()
            runs.append(run)

        return runs

    def new_run(
        self,
        input: Input | dict[str, Any] | BaseModel | str = None,
        name: str | None = None,
        description: str | None = None,
        options: Options | dict[str, str] | None = None,
        configuration: RunConfiguration | dict[str, Any] | None = None,
        json_configurations: dict[str, Any] | None = None,
        input_dir_path: str | None = None,
    ) -> str:
        """
        Run the application locally with the provided input.

        This method is the local equivalent to `cloud.Application.new_run`,
        which submits the input to Nextmv Cloud. This method runs the
        application locally using the `src` of the app.

        Make sure that the `src` attribute is set on the `Application` class
        before running locally, as it is required by the method.

        Parameters
        ----------
        input: Union[Input, dict[str, Any], BaseModel, str]
            Input to use for the run. This can be a `nextmv.Input` object,
            `dict`, `BaseModel` or `str`.

            If `nextmv.Input` is used, and the `input_format` is either
            `nextmv.InputFormat.JSON` or `nextmv.InputFormat.TEXT`, then the
            input data is extracted from the `.data` property.

            If you want to work with `nextmv.InputFormat.CSV_ARCHIVE` or
            `nextmv.InputFormat.MULTI_FILE`, you should use the
            `input_dir_path` argument instead. This argument takes precedence
            over the `input`. If `input_dir_path` is specified, this function
            looks for files in that directory and tars them, to later be
            uploaded using the `upload_large_input` method. If both the
            `input_dir_path` and `input` arguments are provided, the `input`
            is ignored.

            When `input_dir_path` is specified, the `configuration` argument
            must also be provided. More specifically, the
            `RunConfiguration.format.format_input.input_type` parameter
            dictates what kind of input is being submitted to the Nextmv Cloud.
            Make sure that this parameter is specified when working with the
            following input formats:

            - `nextmv.InputFormat.CSV_ARCHIVE`
            - `nextmv.InputFormat.MULTI_FILE`

            When working with JSON or text data, use the `input` argument
            directly.

            In general, if an input is too large, it will be uploaded with the
            `upload_large_input` method.
        name: Optional[str]
            Name of the local run.
        description: Optional[str]
            Description of the local run.
        options: Optional[Union[Options, dict[str, str]]]
            Options to use for the run. This can be a `nextmv.Options` object
            or a dict. If a dict is used, the keys must be strings and the
            values must be strings as well. If a `nextmv.Options` object is
            used, the options are extracted from the `.to_cloud_dict()` method.
            Note that specifying `options` overrides the `input.options` (if
            the `input` is of type `nextmv.Input`).
        configuration: Optional[Union[RunConfiguration, dict[str, Any]]]
            Configuration to use for the run. This can be a
            `cloud.RunConfiguration` object or a dict. If the object is used,
            then the `.to_dict()` method is applied to extract the
            configuration.
        json_configurations: Optional[dict[str, Any]]
            Optional configurations for JSON serialization. This is used to
            customize the serialization before data is sent.
        input_dir_path: Optional[str]
            Path to a directory containing input files. This is useful for
            input formats like `nextmv.InputFormat.CSV_ARCHIVE` or
            `nextmv.InputFormat.MULTI_FILE`. If both `input` and
            `input_dir_path` are specified, the `input` is ignored, and the
            files in the directory are used instead.

        Returns
        -------
        str
            ID (`run_id`) of the local run that was executed.

        Raises
        ------
        ValueError
            If the `src` property for the `Application` is not specified.
            If neither `input` nor `input_dir_path` is specified.
            If `input_dir_path` is specified but `configuration` is not provided.
        FileNotFoundError
            If the manifest.yaml file cannot be found in the specified `src` directory.

        Examples
        --------
        >>> from nextmv.local import Application
        >>> app = Application(id="my-app", src="/path/to/app")
        >>> run_id = app.new_run(
        ...     input={"vehicles": [{"id": "v1"}]},
        ...     options={"duration": "10s"}
        ... )
        >>> print(f"Local run completed with ID: {run_id}")
        """

        configuration = self.__validate_input_dir_path_and_configuration(input_dir_path, configuration)

        if self.src is None:
            raise ValueError("`src` property for the `Application` must be specified to run the application locally")

        if input is None and input_dir_path is None:
            raise ValueError("Either `input` or `input_directory` must be specified")

        try:
            manifest = Manifest.from_yaml(self.src)
        except FileNotFoundError as e:
            raise FileNotFoundError(
                f"Could not find manifest.yaml in {self.src}. Maybe specify a different `src` dir?"
            ) from e

        input_data = None if input_dir_path else self.__extract_input_data(input)
        options_dict = self.__extract_options_dict(options, json_configurations)
        run_config_dict = self.__extract_run_config(input, configuration, input_dir_path)
        run_id = run(
            app_id=self.src,
            src=self.src,
            manifest=manifest,
            run_config=run_config_dict,
            name=name,
            description=description,
            input_data=input_data,
            inputs_dir_path=input_dir_path,
            options=options_dict,
        )

        return run_id

    def new_run_with_result(
        self,
        input: Input | dict[str, Any] | BaseModel | str = None,
        name: str | None = None,
        description: str | None = None,
        run_options: Options | dict[str, str] | None = None,
        polling_options: PollingOptions = DEFAULT_POLLING_OPTIONS,
        configuration: RunConfiguration | dict[str, Any] | None = None,
        json_configurations: dict[str, Any] | None = None,
        input_dir_path: str | None = None,
        output_dir_path: str | None = ".",
    ) -> RunResult:
        """
        Submit an input to start a new local run of the application and poll
        for the result. This is a convenience method that combines the
        `new_run` and `run_result_with_polling` methods, applying
        polling logic to check when the local run succeeded.

        This method is the local equivalent to
        `cloud.Application.new_run_with_result`, which submits the input to
        Nextmv Cloud. This method runs the application locally using the `src`
        of the app.

        Make sure that the `src` attribute is set on the `Application` class
        before running locally, as it is required by the method.

        Parameters
        ----------
        input: Union[Input, dict[str, Any], BaseModel, str]
            Input to use for the run. This can be a `nextmv.Input` object,
            `dict`, `BaseModel` or `str`.

            If `nextmv.Input` is used, and the `input_format` is either
            `nextmv.InputFormat.JSON` or `nextmv.InputFormat.TEXT`, then the
            input data is extracted from the `.data` property.

            If you want to work with `nextmv.InputFormat.CSV_ARCHIVE` or
            `nextmv.InputFormat.MULTI_FILE`, you should use the
            `input_dir_path` argument instead. This argument takes precedence
            over the `input`. If `input_dir_path` is specified, this function
            looks for files in that directory and tars them, to later be
            uploaded using the `upload_large_input` method. If both the
            `input_dir_path` and `input` arguments are provided, the `input` is
            ignored.

            When `input_dir_path` is specified, the `configuration` argument
            must also be provided. More specifically, the
            `RunConfiguration.format.format_input.input_type` parameter
            dictates what kind of input is being submitted to the Nextmv Cloud.
            Make sure that this parameter is specified when working with the
            following input formats:

            - `nextmv.InputFormat.CSV_ARCHIVE`
            - `nextmv.InputFormat.MULTI_FILE`

            When working with JSON or text data, use the `input` argument
            directly.

            In general, if an input is too large, it will be uploaded with the
            `upload_large_input` method.
        name: Optional[str]
            Name of the local run.
        description: Optional[str]
            Description of the local run.
        run_options: Optional[Union[Options, dict[str, str]]]
            Options to use for the run. This can be a `nextmv.Options` object
            or a dict. If a dict is used, the keys must be strings and the
            values must be strings as well. If a `nextmv.Options` object is
            used, the options are extracted from the `.to_cloud_dict()` method.
            Note that specifying `options` overrides the `input.options` (if
            the `input` is of type `nextmv.Input`).
        polling_options: PollingOptions, default=_DEFAULT_POLLING_OPTIONS
            Options to use when polling for the run result.
        configuration: Optional[Union[RunConfiguration, dict[str, Any]]]
            Configuration to use for the run. This can be a
            `cloud.RunConfiguration` object or a dict. If the object is used,
            then the `.to_dict()` method is applied to extract the
            configuration.
        json_configurations: Optional[dict[str, Any]]
            Optional configurations for JSON serialization. This is used to
            customize the serialization before data is sent.
        input_dir_path: Optional[str]
            Path to a directory containing input files. This is useful for
            input formats like `nextmv.InputFormat.CSV_ARCHIVE` or
            `nextmv.InputFormat.MULTI_FILE`. If both `input` and
            `input_dir_path` are specified, the `input` is ignored, and the
            files in the directory are used instead.
        output_dir_path : Optional[str], default="."
            Path to a directory where non-JSON output files will be saved. This
            is required if the output is non-JSON. If the directory does not
            exist, it will be created. Uses the current directory by default.

        Returns
        -------
        RunResult
            Result of the run, including output.

        Raises
        ------
        ValueError
            If the `src` property for the `Application` is not specified. If
            neither `input` nor `inputs_dir_path` is specified. If
            `inputs_dir_path` is specified but `configuration` is not provided.
        FileNotFoundError
            If the manifest.yaml file cannot be found in the specified `src`
            directory.

        Examples
        --------
        >>> from nextmv.local import Application
        >>> app = Application(id="my-app", src="/path/to/app")
        >>> run_result = app.new_run_with_result(
        ...     input={"vehicles": [{"id": "v1"}]},
        ...     options={"duration": "10s"}
        ... )
        >>> print(f"Local run completed with ID: {run_result.id}")
        """

        run_id = self.new_run(
            input=input,
            name=name,
            description=description,
            options=run_options,
            configuration=configuration,
            json_configurations=json_configurations,
            input_dir_path=input_dir_path,
        )

        return self.run_result_with_polling(
            run_id=run_id,
            polling_options=polling_options,
            output_dir_path=output_dir_path,
        )

    def run_logs(self, run_id: str) -> str:
        """
        Get the logs of a local run.

        If the run does not have any logs, or they are empty, then this method
        simply returns a blank string. This method is equivalent to fetching
        the content of the `.nextmv/runs/{run_id}/logs/logs.log` file.

        Parameters
        ----------
        run_id : str
            ID of the run to retrieve logs for.

        Returns
        -------
        str
            The contents of the logs file for the run.

        Raises
        ------
        ValueError
            If the `.nextmv/runs` directory does not exist at the application
            source, or if the specified run ID does not exist.
        """

        runs_dir = os.path.join(self.src, NEXTMV_DIR, RUNS_KEY)
        if not os.path.exists(runs_dir):
            raise ValueError(f"`.nextmv/runs` dir does not exist at app source: {self.src}")

        run_dir = os.path.join(runs_dir, run_id)
        if not os.path.exists(run_dir):
            raise ValueError(f"`{run_id}` run dir does not exist at: {runs_dir}")

        logs_dir = os.path.join(run_dir, LOGS_KEY)
        if not os.path.exists(logs_dir):
            return ""

        logs_file = os.path.join(logs_dir, LOGS_FILE)
        if not os.path.exists(logs_file):
            return ""

        with open(logs_file) as f:
            logs = f.read()

        return logs

    def run_metadata(self, run_id: str) -> RunInformation:
        """
        Get the metadata of a local run.

        This method is the local equivalent to
        `cloud.Application.run_metadata`, which retrieves the metadata of a
        remote run in Nextmv Cloud. This method is used to get the metadata of
        a run that was executed locally using the `new_run` or
        `new_run_with_result` method.

        Retrieves information about a run without including the run output.
        This is useful when you only need the run's status and metadata.

        Parameters
        ----------
        run_id : str
            ID of the run to retrieve metadata for.

        Returns
        -------
        RunInformation
            Metadata of the run (run information without output).

        Raises
        ------
        ValueError
            If the `.nextmv/runs` directory does not exist at the application
            source, or if the specified run ID does not exist.

        Examples
        --------
        >>> metadata = app.run_metadata("run-789")
        >>> print(metadata.metadata.status_v2)
        StatusV2.succeeded
        """

        runs_dir = os.path.join(self.src, NEXTMV_DIR, RUNS_KEY)
        if not os.path.exists(runs_dir):
            raise ValueError(f"`.nextmv/runs` dir does not exist at app source: {self.src}")

        run_dir = os.path.join(runs_dir, run_id)
        if not os.path.exists(run_dir):
            raise ValueError(f"`{run_id}` run dir does not exist at: {runs_dir}")

        info_file = os.path.join(run_dir, f"{run_id}.json")
        if not os.path.exists(info_file):
            raise ValueError(f"`{info_file}` file does not exist at: {run_dir}")

        with open(info_file) as f:
            info_dict = json.load(f)

        info = RunInformation.from_dict(info_dict)

        return info

    def run_result(self, run_id: str, output_dir_path: str | None = ".") -> RunResult:
        """
        Get the local result of a run.

        This method is the local equivalent to `cloud.Application.run_result`,
        which retrieves the result of a remote run in Nextmv Cloud. This method
        is used to get the result of a run that was executed locally using the
        `new_run` or `new_run_with_result` method.

        Retrieves the complete result of a run, including the run output.

        Parameters
        ----------
        run_id : str
            ID of the run to get results for.
        output_dir_path : Optional[str], default="."
            Path to a directory where non-JSON output files will be saved. This
            is required if the output is non-JSON. If the directory does not
            exist, it will be created. Uses the current directory by default.

        Returns
        -------
        RunResult
            Result of the run, including output.

        Raises
        ------
        ValueError
            If the `.nextmv/runs` directory does not exist at the application
            source, or if the specified run ID does not exist.

        Examples
        --------
        >>> result = app.run_result("run-123")
        >>> print(result.metadata.status_v2)
        'succeeded'
        """

        run_information = self.run_metadata(run_id=run_id)

        return self.__run_result(
            run_id=run_id,
            run_information=run_information,
            output_dir_path=output_dir_path,
        )

    def run_result_with_polling(
        self,
        run_id: str,
        polling_options: PollingOptions = DEFAULT_POLLING_OPTIONS,
        output_dir_path: str | None = ".",
    ) -> RunResult:
        """
        Get the result of a local run with polling.

        This method is the local equivalent to
        `cloud.Application.run_result_with_polling`, which retrieves the result
        of a remote run in Nextmv Cloud. This method is used to get the result
        of a run that was executed locally using the `new_run` or
        `new_run_with_result` method.

        Retrieves the result of a run including the run output. This method
        polls for the result until the run finishes executing or the polling
        strategy is exhausted.

        Parameters
        ----------
        run_id : str
            ID of the run to retrieve the result for.
        polling_options : PollingOptions, default=_DEFAULT_POLLING_OPTIONS
            Options to use when polling for the run result.
        output_dir_path : Optional[str], default="."
            Path to a directory where non-JSON output files will be saved. This
            is required if the output is non-JSON. If the directory does not
            exist, it will be created. Uses the current directory by default.

        Returns
        -------
        RunResult
            Complete result of the run including output data.

        Raises
        ------
        requests.HTTPError
            If the response status code is not 2xx.
        TimeoutError
            If the run does not complete after the polling strategy is
            exhausted based on time duration.
        RuntimeError
            If the run does not complete after the polling strategy is
            exhausted based on number of tries.

        Examples
        --------
        >>> from nextmv.cloud import PollingOptions
        >>> # Create custom polling options
        >>> polling_opts = PollingOptions(max_tries=50, max_duration=600)
        >>> # Get run result with polling
        >>> result = app.run_result_with_polling("run-123", polling_opts)
        >>> print(result.output)
        {'solution': {...}}
        """

        def polling_func() -> tuple[Any, bool]:
            run_information = self.run_metadata(run_id=run_id)
            if run_information.metadata.status_v2 in {
                StatusV2.succeeded,
                StatusV2.failed,
                StatusV2.canceled,
            }:
                return run_information, True

            return None, False

        run_information = poll(polling_options=polling_options, polling_func=polling_func)

        return self.__run_result(
            run_id=run_id,
            run_information=run_information,
            output_dir_path=output_dir_path,
        )

    def run_visuals(self, run_id: str) -> None:
        """
        Open the local run visuals in a web browser.

        This method opens the visual representation of a locally executed run
        in the default web browser. It assumes that the run was executed locally
        using the `new_run` or `new_run_with_result` method and that
        the necessary visualization files are present.

        If the run was correctly configured to produce visual assets, then the
        run will contain a `visuals` directory with one or more HTML files.
        Each file is opened in a new tab in the default web browser.

        Parameters
        ----------
        run_id : str
            ID of the local run to visualize.

        Raises
        ------
        ValueError
            If the `.nextmv/runs` directory does not exist at the application
            source, or if the specified run ID does not exist.
        """

        runs_dir = os.path.join(self.src, NEXTMV_DIR, RUNS_KEY)
        if not os.path.exists(runs_dir):
            raise ValueError(f"`.nextmv/runs` dir does not exist at app source: {self.src}")

        run_dir = os.path.join(runs_dir, run_id)
        if not os.path.exists(run_dir):
            raise ValueError(f"`{run_id}` run dir does not exist at: {runs_dir}")

        visuals_dir = os.path.join(run_dir, "visuals")
        if not os.path.exists(visuals_dir):
            raise ValueError(f"`visuals` dir does not exist at: {run_dir}")

        for file in os.listdir(visuals_dir):
            if file.endswith(".html"):
                file_path = os.path.join(visuals_dir, file)
                webbrowser.open_new_tab(f"file://{os.path.realpath(file_path)}")

    def sync(  # noqa: C901
        self,
        target: cloud.Application,
        run_ids: list[str] | None = None,
        instance_id: str | None = None,
        verbose: bool | None = False,
    ) -> None:
        """
        Sync the local application to a Nextmv Cloud application target.

        The `Application` class allows you to perform and handle local
        application runs with methods such as:

        - `new_run`
        - `new_run_with_result`
        - `run_metadata`
        - `run_result`
        - `run_result_with_polling`

        The runs produced locally live under `self.src/.nextmv/runs`. This
        method syncs those runs to a Nextmv Cloud application target, making
        them available for remote execution and management.

        Parameters
        ----------
        target : cloud.Application
            Target Nextmv Cloud application where the local application runs
            will be synced to.
        run_ids : Optional[list[str]], default=None
            List of run IDs to sync. If None, all local runs found under
            `self.src/.nextmv/runs` will be synced.
        instance_id : Optional[str], default=None
            Optional instance ID if you want to associate your runs with an
            instance.
        verbose : Optional[bool], default=False
            Whether to print verbose output during the sync process. Useful for
            debugging a large number of runs being synced.

        Raises
        ------
        ValueError
            If the `src` property is not specified.
        ValueError
            If the `client` property is not specified.
        ValueError
            If the application does not exist in Nextmv Cloud.
        ValueError
            If a run does not exist locally.
        requests.HTTPError
            If the response status code is not 2xx.
        """
        if self.src is None:
            raise ValueError(
                "`src` property for the `Application` must be specified to sync the application to Nextmv Cloud"
            )

        if target.client is None:
            raise ValueError(
                "`client` property for the target `Application` must be specified to sync the application to Cloud"
            )

        if not target.exists(target.client, target.id):
            raise ValueError(
                "target Application does not exist in Nextmv Cloud, create it with `cloud.Application.new`"
            )
        if verbose:
            log(f"☁️ Starting sync of local application `{self.src}` to Nextmv Cloud application `{target.id}`.")

        # Create a temp dir to store the outputs that are written by default to
        # ".". During the sync process, we don't need to keep these outputs, so
        # we can use a temp dir that will be deleted after the sync is done.
        with tempfile.TemporaryDirectory(prefix="nextmv-sync-run-") as temp_results_dir:
            runs_dir = os.path.join(self.src, NEXTMV_DIR, RUNS_KEY)
            if run_ids is None:
                # If runs are not specified, by default we sync all local runs that
                # can be found.
                dirs = os.listdir(runs_dir)
                run_ids = [d for d in dirs if os.path.isdir(os.path.join(runs_dir, d))]

                if verbose:
                    log(f"ℹ️  Found {len(run_ids)} local runs to sync from {runs_dir}.")
            else:
                if verbose:
                    log(f"ℹ️  Syncing {len(run_ids)} specified local runs from {runs_dir}.")

            total = 0
            for run_id in run_ids:
                synced = self.__sync_run(
                    target=target,
                    run_id=run_id,
                    runs_dir=runs_dir,
                    temp_dir=temp_results_dir,
                    instance_id=instance_id,
                    verbose=verbose,
                )
                if synced:
                    total += 1

            if verbose:
                log(
                    f"🚀 Process completed, synced local application `{self.src}` to "
                    f"Nextmv Cloud application `{target.id}`: "
                    f"{total}/{len(run_ids)} runs."
                )

    def __run_result(
        self,
        run_id: str,
        run_information: RunInformation,
        output_dir_path: str | None = ".",
    ) -> RunResult:
        """
        Get the result of a local run.

        This is a private method that retrieves the complete result of a run,
        including the output data, from a local source. This method serves as
        the base implementation for retrieving run results, regardless of
        polling strategy.

        Parameters
        ----------
        run_id : str
            ID of the run to retrieve the result for.
        run_information : RunInformation
            Information about the run, including metadata such as output size.
        output_dir_path : Optional[str], default="."
            Path to a directory where non-JSON output files will be saved. This
            is required if the output is non-JSON. If the directory does not
            exist, it will be created. Uses the current directory by default.

        Returns
        -------
        RunResult
            Result of the run, including all metadata and output data.

        Raises
        ------
        ValueError
            If the output format is not JSON and no output_dir_path is
            provided.
            If the output format is unknown.
        """

        result = RunResult.from_dict(run_information.to_dict())
        if result.metadata.error:
            result.error_log = ErrorLog(error=result.metadata.error)

        if result.metadata.status_v2 != StatusV2.succeeded:
            return result

        # See whether we can attach the output directly or need to save to the given
        # directory
        output_type = run_information.metadata.format.format_output.output_type
        if output_type != OutputFormat.JSON and (not output_dir_path or output_dir_path == ""):
            raise ValueError(
                "The output format is not JSON: an `output_dir_path` must be provided.",
            )

        runs_dir = os.path.join(self.src, NEXTMV_DIR, RUNS_KEY)
        solutions_dir = os.path.join(runs_dir, run_id, OUTPUTS_KEY, SOLUTIONS_KEY)

        if output_type == OutputFormat.JSON:
            with open(os.path.join(solutions_dir, DEFAULT_OUTPUT_JSON_FILE)) as f:
                result.output = json.load(f)
        elif output_type in {OutputFormat.CSV_ARCHIVE, OutputFormat.MULTI_FILE}:
            shutil.copytree(solutions_dir, output_dir_path, dirs_exist_ok=True)
        else:
            raise ValueError(f"Unknown output type: {output_type}")

        return result

    def __validate_input_dir_path_and_configuration(
        self,
        input_dir_path: str | None,
        configuration: RunConfiguration | dict[str, Any] | None,
    ) -> RunConfiguration:
        """
        Auxiliary function to validate the directory path and configuration.
        """

        if configuration is None:
            if self.manifest.configuration is not None and self.manifest.configuration.content is not None:
                configuration = RunConfiguration(
                    format=Format(
                        format_input=FormatInput(
                            input_type=self.manifest.configuration.content.format,
                        ),
                    ),
                )
        elif isinstance(configuration, dict):
            # Forcefully turn the configuration into a RunConfiguration object to
            # make it easier to deal with in the other functions.
            configuration = RunConfiguration.from_dict(configuration)

        if input_dir_path is None or input_dir_path == "":
            return configuration

        if configuration is None:
            raise ValueError(
                "If `dir_path` is provided, either a `RunConfiguration` must also be provided or "
                "the application's manifest (app.yaml) must include the format under "
                "`configuration.content.format`.",
            )

        config_format = configuration.format
        if config_format is None:
            raise ValueError(
                "If `dir_path` is provided, `RunConfiguration.format` must also be provided.",
            )

        input_type = config_format.format_input
        if input_type is None:
            raise ValueError(
                "If `dir_path` is provided, `RunConfiguration.format.format_input` must also be provided.",
            )

        if input_type is None or input_type in (InputFormat.JSON, InputFormat.TEXT):
            raise ValueError(
                "If `dir_path` is provided, `RunConfiguration.format.format_input.input_type` must be set to "
                f"a valid type. Valid types are: {[InputFormat.CSV_ARCHIVE, InputFormat.MULTI_FILE]}",
            )

        return configuration

    def __extract_input_data(
        self,
        input: Input | dict[str, Any] | BaseModel | str = None,
    ) -> dict[str, Any] | str | None:
        """
        Auxiliary function to extract the input data from the input, based on
        its type.
        """

        input_data = None
        if isinstance(input, BaseModel):
            input_data = input.to_dict()
        elif isinstance(input, dict) or isinstance(input, str):
            input_data = input
        elif isinstance(input, Input):
            input_data = input.data

        return input_data

    def __extract_options_dict(
        self,
        options: Options | dict[str, str] | None = None,
        json_configurations: dict[str, Any] | None = None,
    ) -> dict[str, str]:
        """
        Auxiliary function to extract the options that will be sent to the
        application for execution.
        """

        options_dict = {}
        if options is not None:
            if isinstance(options, Options):
                options_dict = options.to_dict_cloud()
            elif isinstance(options, dict):
                for k, v in options.items():
                    if isinstance(v, str):
                        options_dict[k] = v
                    else:
                        options_dict[k] = deflated_serialize_json(v, json_configurations=json_configurations)

        return options_dict

    def __extract_run_config(
        self,
        input: Input | dict[str, Any] | BaseModel | str = None,
        configuration: RunConfiguration | dict[str, Any] | None = None,
        dir_path: str | None = None,
    ) -> dict[str, Any]:
        """
        Auxiliary function to extract the run configuration that will be sent
        to the application for execution.
        """

        if configuration is not None:
            configuration_dict = (
                configuration.to_dict() if isinstance(configuration, RunConfiguration) else configuration
            )
            return configuration_dict

        configuration = RunConfiguration()
        configuration.resolve(input=input, dir_path=dir_path)
        configuration_dict = configuration.to_dict()

        return configuration_dict

    def __sync_run(  # noqa: C901
        self,
        target: cloud.Application,
        run_id: str,
        runs_dir: str,
        temp_dir: str,
        instance_id: str | None = None,
        verbose: bool | None = False,
    ) -> bool:
        """
        Syncs a local run to a Nextmv Cloud target application. Returns True if
        the run was synced, False if it was skipped (already synced).
        """

        if verbose:
            log(f"🔄 Syncing local run `{run_id}`... ")

        # For files-based runs, the result files are written by default to ".".
        # Avoid this using a dedicated temp dir.
        run_result = self.run_result(run_id, output_dir_path=temp_dir)
        input_type = run_result.metadata.format.format_input.input_type

        # Skip runs that have already been synced.
        synced_run, already_synced = run_result.is_synced(app_id=target.id, instance_id=instance_id)
        if already_synced:
            if verbose:
                log(f"   ⏭️  Skipping local run `{run_id}`, already synced with {synced_run.to_dict()}.")

            return False

        # Check that it is a valid run with inputs, outputs, logs, etc.
        if not self.__valid_run_result(run_result, runs_dir, run_id):
            if verbose:
                log(f"   ❌  Skipping local run `{run_id}`, invalid run (missing inputs, outputs or logs).")

            return False

        status = TrackedRunStatus.SUCCEEDED
        if run_result.metadata.status_v2 != StatusV2.succeeded:
            status = TrackedRunStatus.FAILED

        # Read the logs of the run and place each line as an element in a list
        run_dir = os.path.join(runs_dir, run_id)
        with open(os.path.join(run_dir, LOGS_KEY, LOGS_FILE)) as f:
            stderr_logs = f.read()

        # Create the tracked run object and start configuring it.
        tracked_run = TrackedRun(
            status=status,
            duration=int(run_result.metadata.duration),
            error=run_result.metadata.error,
            logs=stderr_logs,
            name=run_result.name,
            description=run_result.description,
        )

        # Resolve the input according to its type.
        inputs_path = os.path.join(run_dir, INPUTS_KEY)
        if input_type == InputFormat.JSON:
            with open(os.path.join(inputs_path, DEFAULT_INPUT_JSON_FILE)) as f:
                tracked_run.input = json.load(f)
        elif input_type == InputFormat.TEXT:
            with open(os.path.join(inputs_path, "input")) as f:
                tracked_run.input = f.read()
        else:
            tracked_run.input_dir_path = inputs_path

        # Resolve the output according to its type.
        output_type = run_result.metadata.format.format_output.output_type
        if output_type == OutputFormat.JSON:
            tracked_run.output = run_result.output
        else:
            tracked_run.output_dir_path = os.path.join(run_dir, OUTPUTS_KEY, SOLUTIONS_KEY)

        # Resolve the statistics according to their type and presence. If
        # working with JSON, the statistics should be resolved from the output.
        if output_type in {OutputFormat.CSV_ARCHIVE, OutputFormat.MULTI_FILE}:
            stats_file_path = os.path.join(run_dir, OUTPUTS_KEY, STATISTICS_KEY, f"{STATISTICS_KEY}.json")
            if os.path.exists(stats_file_path):
                with open(stats_file_path) as f:
                    tracked_run.statistics = json.load(f)

        # Resolve the assets according to their type and presence. If working
        # with JSON, the assets should be resolved from the output.
        if output_type in {OutputFormat.CSV_ARCHIVE, OutputFormat.MULTI_FILE}:
            assets_file_path = os.path.join(run_dir, OUTPUTS_KEY, ASSETS_KEY, f"{ASSETS_KEY}.json")
            if os.path.exists(assets_file_path):
                with open(assets_file_path) as f:
                    tracked_run.assets = json.load(f)

        # Actually sync the run by tracking it remotely on Nextmv Cloud.
        configuration = RunConfiguration(
            format=Format(
                format_input=run_result.metadata.format.format_input,
                format_output=run_result.metadata.format.format_output,
            ),
        )
        tracked_id = target.track_run(
            tracked_run=tracked_run,
            instance_id=instance_id,
            configuration=configuration,
        )

        # Mark the local run as synced by updating the local run info.
        synced_run = SyncedRun(
            run_id=tracked_id,
            synced_at=datetime.now(timezone.utc),
            app_id=target.id,
            instance_id=instance_id,
        )
        run_result.add_synced_run(synced_run)
        with open(os.path.join(run_dir, f"{run_id}.json"), "w") as f:
            json.dump(run_result.to_dict(), f, indent=2)

        if verbose:
            log(f"✅ Synced local run `{run_id}` as remote run `{synced_run.to_dict()}`.")

        return True

    def __valid_run_result(self, run_result: RunResult, runs_dir: str, run_id: str) -> bool:
        """
        Validate that a run result has all required files and directories.

        This method checks that a local run has the expected directory structure
        and files, including inputs, outputs, and logs.

        Parameters
        ----------
        run_result : RunResult
            The run result to validate.
        runs_dir : str
            Path to the runs directory.
        run_id : str
            ID of the run to validate.

        Returns
        -------
        bool
            True if the run is valid, False otherwise.
        """
        run_dir = os.path.join(runs_dir, run_id)

        # Check that the run directory exists
        if not os.path.exists(run_dir):
            return False

        # Validate inputs
        if not self.__validate_inputs(run_dir, run_result.metadata.format.format_input.input_type):
            return False

        # Validate outputs
        format_output = run_result.metadata.format.format_output
        if format_output is None or not format_output:
            return False

        output_type = format_output.output_type
        if output_type is None or output_type == "":
            return False

        if not self.__validate_outputs(run_dir, output_type):
            return False

        # Validate logs
        if not self.__validate_logs(run_dir):
            return False

        return True

    def __validate_inputs(self, run_dir: str, input_type: InputFormat) -> bool:
        """Validate that the inputs directory and files exist for the given input type."""
        inputs_path = os.path.join(run_dir, INPUTS_KEY)
        if not os.path.exists(inputs_path):
            return False

        if input_type == InputFormat.JSON:
            input_file = os.path.join(inputs_path, DEFAULT_INPUT_JSON_FILE)

            return os.path.isfile(input_file)

        if input_type == InputFormat.TEXT:
            input_file = os.path.join(inputs_path, "input")

            return os.path.isfile(input_file)

        # For CSV_ARCHIVE and MULTI_FILE, inputs_path should be a directory
        return os.path.isdir(inputs_path)

    def __validate_outputs(self, run_dir: str, output_type: OutputFormat) -> bool:
        """Validate that the outputs directory and files exist for the given output type."""
        outputs_dir = os.path.join(run_dir, OUTPUTS_KEY)
        if not os.path.exists(outputs_dir):
            return False

        solutions_dir = os.path.join(outputs_dir, SOLUTIONS_KEY)
        if not os.path.exists(solutions_dir):
            return False

        if output_type == OutputFormat.JSON:
            solution_file = os.path.join(solutions_dir, DEFAULT_OUTPUT_JSON_FILE)

            return os.path.isfile(solution_file)

        # For CSV_ARCHIVE and MULTI_FILE, solutions_dir should be a directory
        return os.path.isdir(solutions_dir)

    def __validate_logs(self, run_dir: str) -> bool:
        """Validate that the logs directory and file exist."""
        logs_dir = os.path.join(run_dir, LOGS_KEY)
        if not os.path.exists(logs_dir):
            return False

        logs_file = os.path.join(logs_dir, LOGS_FILE)

        return os.path.isfile(logs_file)



import os
import platform
import shutil
import subprocess
import tempfile
import unittest

from nextmv.cloud.package import _get_shell_command_elements, _package
from nextmv.manifest import Manifest, ManifestType


class TestPackageOneFile(unittest.TestCase):
    def setUp(self):
        self.app_dir = tempfile.mkdtemp()
        self.manifest = Manifest(type=ManifestType.PYTHON, files=["main.py"], python=None)
        with open(os.path.join(self.app_dir, "main.py"), "w") as f:
            f.write("print('Hello, World!')")

    def tearDown(self):
        shutil.rmtree(self.app_dir)

    def test_package_creates_tarball(self):
        tar_file, output_dir = _package(self.app_dir, self.manifest, verbose=False)
        self.assertTrue(os.path.isfile(tar_file))
        self.assertTrue(os.path.isdir(output_dir))

    def test_package_missing_files(self):
        self.manifest.files.append("missing_file.py")
        with self.assertRaises(Exception) as context:
            _package(self.app_dir, self.manifest, verbose=False)
        self.assertIn("could not find files listed in manifest", str(context.exception))

    def test_package_mandatory_files(self):
        self.manifest.type = ManifestType.GO
        with self.assertRaises(Exception) as context:
            _package(self.app_dir, self.manifest, verbose=False)
        self.assertIn("missing mandatory files", str(context.exception))


class TestPackageDir(unittest.TestCase):
    def setUp(self):
        self.app_dir = tempfile.mkdtemp()
        self.manifest = Manifest(type=ManifestType.PYTHON, files=["main.py", "app/"], python=None)
        with open(os.path.join(self.app_dir, "main.py"), "w") as f:
            f.write("print('Hello, World!')")

        os.makedirs(os.path.join(self.app_dir, "app"))
        with open(os.path.join(self.app_dir, "app", "main.py"), "w") as f:
            f.write("print('Hello, World!')")

    def tearDown(self):
        shutil.rmtree(self.app_dir)

    def test_package_creates_tarball(self):
        tar_file, output_dir = _package(self.app_dir, self.manifest, verbose=False)
        self.assertTrue(os.path.isfile(tar_file))
        self.assertTrue(os.path.isdir(output_dir))

        with tempfile.TemporaryDirectory() as tmpdir:
            shutil.unpack_archive(tar_file, tmpdir)
            self.assertTrue(os.path.isfile(os.path.join(tmpdir, "main.py")))
            self.assertTrue(os.path.isdir(os.path.join(tmpdir, "app")))
            self.assertTrue(os.path.isfile(os.path.join(tmpdir, "app", "main.py")))

    def test_package_missing_files(self):
        self.manifest.files.append("missing_file.py")
        with self.assertRaises(Exception) as context:
            _package(self.app_dir, self.manifest, verbose=False)
        self.assertIn("could not find files listed in manifest", str(context.exception))

    def test_package_mandatory_files(self):
        self.manifest.type = ManifestType.GO
        with self.assertRaises(Exception) as context:
            _package(self.app_dir, self.manifest, verbose=False)
        self.assertIn("missing mandatory files", str(context.exception))

    def test_get_shell_command_elements(self):
        if platform.system() == "Windows":
            command = _get_shell_command_elements("echo Hello World")
            subprocess.run(command, check=True)
        else:
            command = _get_shell_command_elements("echo 'Hello World'")
            subprocess.run(command, check=True)



"""This module contains definitions for scenario tests.

Classes
-------
ScenarioConfiguration
    Configuration for a scenario with multiple option values.
ScenarioInputType
    Enumeration of input types for a scenario.
ScenarioInput
    Input to be processed in a scenario.
Scenario
    A test case for comparing a decision model with inputs and configurations.
"""

import itertools
from dataclasses import dataclass
from enum import Enum
from typing import Any


@dataclass
class ScenarioConfiguration:
    """
    Configuration for a scenario.

    You can import the `ScenarioConfiguration` class directly from `cloud`:

    ```python
    from nextmv.cloud import ScenarioConfiguration
    ```

    You can define multiple values for a single option, which will result in
    multiple runs being created. For example, if you have a configuration
    option "x" with values [1, 2], and a configuration option "y" with values
    [3, 4], then the following runs will be created:
    - x=1, y=3
    - x=1, y=4
    - x=2, y=3
    - x=2, y=4

    Parameters
    ----------
    name : str
        Name of the configuration option.
    values : list[str]
        List of values for the configuration option.

    Examples
    --------
    >>> from nextmv.cloud import ScenarioConfiguration
    >>> config = ScenarioConfiguration(name="solver", values=["simplex", "interior-point"])
    """

    name: str
    """Name of the configuration option."""
    values: list[str]
    """List of values for the configuration option."""

    def __post_init__(self):
        """
        Post-initialization method to ensure that the values are unique.

        Raises
        ------
        ValueError
            If the configuration values list is empty.
        """
        if len(self.values) <= 0:
            raise ValueError("Configuration values must be non-empty.")


class ScenarioInputType(str, Enum):
    """
    Type of input for a scenario. This is used to determine how the input
    should be processed.

    You can import the `ScenarioInputType` class directly from `cloud`:

    ```python
    from nextmv.cloud import ScenarioInputType
    ```

    Parameters
    ----------
    INPUT_SET : str
        The data in the scenario is an input set.
    INPUT : str
        The data in the scenario is an input.
    NEW : str
        The data in the scenario is new data.

    Examples
    --------
    >>> from nextmv.cloud import ScenarioInputType
    >>> input_type = ScenarioInputType.INPUT_SET
    """

    INPUT_SET = "input_set"
    """The data in the scenario is an input set."""
    INPUT = "input"
    """The data in the scenario is an input."""
    NEW = "new"
    """The data in the scenario is new data."""


@dataclass
class ScenarioInput:
    """
    Input to be processed in a scenario.

    You can import the `ScenarioInput` class directly from `cloud`:

    ```python
    from nextmv.cloud import ScenarioInput
    ```

    The type of input is determined by the `scenario_input_type` attribute.
    The input can be a single input set ID, a list of input IDs, or raw data.
    The data itself of the scenario input is tracked by the `scenario_input_data`
    attribute.

    Parameters
    ----------
    scenario_input_type : ScenarioInputType
        Type of input for the scenario. This is used to determine how the input
        should be processed.
    scenario_input_data : Union[str, list[str], list[dict[str, Any]]]
        Input data for the scenario. This can be a single input set ID
        (`str`), a list of input IDs (`list[str]`), or raw data
        (`list[dict[str, Any]]`). If you provide a `list[str]` (list of
        inputs), a new input set will be created using these inputs. A similar
        behavior occurs when providing raw data (`list[dict[str, Any]]`). All
        the entries in the list of raw dicts will be collected to create a new
        input set.

    Examples
    --------
    >>> from nextmv.cloud import ScenarioInput, ScenarioInputType
    >>> # Using an existing input set
    >>> input_set = ScenarioInput(
    ...     scenario_input_type=ScenarioInputType.INPUT_SET,
    ...     scenario_input_data="input-set-id-123"
    ... )
    >>> # Using a list of inputs
    >>> inputs = ScenarioInput(
    ...     scenario_input_type=ScenarioInputType.INPUT,
    ...     scenario_input_data=["input-id-1", "input-id-2"]
    ... )
    >>> # Using raw data
    >>> raw_data = ScenarioInput(
    ...     scenario_input_type=ScenarioInputType.NEW,
    ...     scenario_input_data=[{"id": 1, "value": "data1"}, {"id": 2, "value": "data2"}]
    ... )
    """

    scenario_input_type: ScenarioInputType
    """
    Type of input for the scenario. This is used to determine how the input
    should be processed.
    """
    scenario_input_data: str | list[str] | list[dict[str, Any]]
    """
    Input data for the scenario. This can be a single input set ID (`str`), a
    list of input IDs (`list[str]`), or raw data (`list[dict[str, Any]]`).
    """

    def __post_init__(self):
        """
        Post-initialization method to ensure that the input data is valid.

        Raises
        ------
        ValueError
            If the scenario input type and data type don't match:
            - When using INPUT_SET, scenario_input_data must be a string
            - When using INPUT or NEW, scenario_input_data must be a list
        """
        if self.scenario_input_type == ScenarioInputType.INPUT_SET and not isinstance(self.scenario_input_data, str):
            raise ValueError("Scenario input type must be a string when using an input set.")
        elif self.scenario_input_type == ScenarioInputType.INPUT and not isinstance(self.scenario_input_data, list):
            raise ValueError("Scenario input type must be a list when using an input.")
        elif self.scenario_input_type == ScenarioInputType.NEW and not isinstance(self.scenario_input_data, list):
            raise ValueError("Scenario input type must be a list when using new data.")


@dataclass
class Scenario:
    """
    A scenario is a test case that is used to compare a decision model being
    executed with a set of inputs and configurations.

    You can import the `Scenario` class directly from `cloud`:

    ```python
    from nextmv.cloud import Scenario
    ```

    A scenario encapsulates all the necessary information to run a test case
    against a decision model. Each scenario includes input data, an instance ID,
    and can optionally include configuration options that define different
    variations of the run.

    Parameters
    ----------
    scenario_input : ScenarioInput
        Input for the scenario. The input is composed of a type and data. Make
        sure you use the `ScenarioInput` class to create the input.
    instance_id : str
        ID of the instance to be used for the scenario.
    scenario_id : Optional[str]
        Optional ID of the scenario. The default value will be set as
        `scenario-<index>` if not set.
    configuration : Optional[list[ScenarioConfiguration]]
        Optional configuration for the scenario. Use this attribute to
        configure variation of options for the scenario.

    Examples
    --------
    >>> from nextmv.cloud import Scenario, ScenarioInput, ScenarioInputType, ScenarioConfiguration
    >>> # Creating a simple scenario with an input set
    >>> input_data = ScenarioInput(
    ...     scenario_input_type=ScenarioInputType.INPUT_SET,
    ...     scenario_input_data="input-set-id-123"
    ... )
    >>> scenario = Scenario(
    ...     scenario_input=input_data,
    ...     instance_id="instance-id-456",
    ...     scenario_id="my-test-scenario"
    ... )
    >>>
    >>> # Creating a scenario with configuration options
    >>> config_options = [
    ...     ScenarioConfiguration(name="solver", values=["simplex", "interior-point"]),
    ...     ScenarioConfiguration(name="timeout", values=["10", "30", "60"])
    ... ]
    >>> scenario_with_config = Scenario(
    ...     scenario_input=input_data,
    ...     instance_id="instance-id-456",
    ...     configuration=config_options
    ... )
    """

    scenario_input: ScenarioInput
    """
    Input for the scenario. The input is composed of a type and data. Make sure
    you use the `ScenarioInput` class to create the input.
    """
    instance_id: str
    """ID of the instance to be used for the scenario."""

    scenario_id: str | None = None
    """
    Optional ID of the scenario. The default value will be set as
    `scenario-<index>` if not set.
    """
    configuration: list[ScenarioConfiguration] | None = None
    """Optional configuration for the scenario. Use this attribute to configure
    variation of options for the scenario.
    """

    def option_combinations(self) -> list[dict[str, str]]:
        """
        Creates the combination of options that are derived from the
        `configuration` property.

        This method calculates the cross-product of all configuration
        options to generate all possible combinations. If no configuration
        is provided, it returns a list with an empty dictionary.

        Returns
        -------
        list[dict[str, str]]
            A list of dictionaries where each dictionary represents a set of
            options derived from the configuration.

        Examples
        --------
        >>> from nextmv.cloud import Scenario, ScenarioInput, ScenarioInputType, ScenarioConfiguration
        >>> input_data = ScenarioInput(
        ...     scenario_input_type=ScenarioInputType.INPUT_SET,
        ...     scenario_input_data="input-set-id"
        ... )
        >>> config = [
        ...     ScenarioConfiguration(name="x", values=["1", "2"]),
        ...     ScenarioConfiguration(name="y", values=["3", "4"])
        ... ]
        >>> scenario = Scenario(
        ...     scenario_input=input_data,
        ...     instance_id="instance-id",
        ...     configuration=config
        ... )
        >>> scenario.option_combinations()
        [{'x': '1', 'y': '3'}, {'x': '1', 'y': '4'}, {'x': '2', 'y': '3'}, {'x': '2', 'y': '4'}]
        """

        if self.configuration is None or len(self.configuration) == 0:
            return [{}]

        keys, value_lists = zip(*((config.name, config.values) for config in self.configuration), strict=False)
        combinations = [dict(zip(keys, values, strict=False)) for values in itertools.product(*value_lists)]

        return combinations


def _option_sets(scenarios: list[Scenario]) -> dict[str, dict[str, dict[str, str]]]:
    """
    Creates options sets that are derived from `scenarios`.

    The options sets are grouped by scenario ID. The cross-product of the
    configuration options is created to generate all possible combinations
    of options. Each combination is given a unique key based on the scenario ID
    and a combination index.

    Parameters
    ----------
    scenarios : list[Scenario]
        List of scenarios to be tested.

    Returns
    -------
    dict[str, dict[str, dict[str, str]]]
        A dictionary where the keys are scenario IDs and the values are
        dictionaries of option sets. Each option set is a dictionary where the
        keys are option names and the values are the corresponding option
        values.

    Examples
    --------
    >>> from nextmv.cloud import Scenario, ScenarioInput, ScenarioInputType, ScenarioConfiguration
    >>> input_data = ScenarioInput(
    ...     scenario_input_type=ScenarioInputType.INPUT_SET,
    ...     scenario_input_data="input-set-id"
    ... )
    >>> config = [ScenarioConfiguration(name="x", values=["1", "2"])]
    >>> scenario = Scenario(
    ...     scenario_input=input_data,
    ...     instance_id="instance-id",
    ...     scenario_id="test-scenario",
    ...     configuration=config
    ... )
    >>> _option_sets([scenario])
    {'test-scenario': {'test-scenario_0': {'x': '1'}, 'test-scenario_1': {'x': '2'}}}
    """

    sets_by_scenario = {}
    scenarios_by_id = _scenarios_by_id(scenarios)
    for scenario_id, scenario in scenarios_by_id.items():
        combinations = scenario.option_combinations()
        option_sets = {}
        for comb_ix, combination in enumerate(combinations):
            option_sets[f"{scenario_id}_{comb_ix}"] = combination

        sets_by_scenario[scenario_id] = option_sets

    return sets_by_scenario


def _scenarios_by_id(scenarios: list[Scenario]) -> dict[str, Scenario]:
    """
    Maps scenarios to their IDs.

    This function builds a dictionary that maps each scenario to its ID.
    If a scenario doesn't have an ID defined, one is created using the format
    "scenario-{index}". The function also checks that there are no duplicate
    scenario IDs in the provided list.

    Parameters
    ----------
    scenarios : list[Scenario]
        List of scenarios to be mapped.

    Returns
    -------
    dict[str, Scenario]
        A dictionary where keys are scenario IDs and values are the corresponding
        Scenario objects.

    Raises
    ------
    ValueError
        If duplicate scenario IDs are found in the list.

    Examples
    --------
    >>> from nextmv.cloud import Scenario, ScenarioInput, ScenarioInputType
    >>> input_data = ScenarioInput(
    ...     scenario_input_type=ScenarioInputType.INPUT_SET,
    ...     scenario_input_data="input-set-id"
    ... )
    >>> scenarios = [
    ...     Scenario(scenario_input=input_data, instance_id="instance-1", scenario_id="test-1"),
    ...     Scenario(scenario_input=input_data, instance_id="instance-2")
    ... ]
    >>> result = _scenarios_by_id(scenarios)
    >>> sorted(list(result.keys()))
    ['scenario-2', 'test-1']
    """

    scenario_by_id = {}
    ids_used = {}
    for scenario_ix, scenario in enumerate(scenarios, start=1):
        scenario_id = f"scenario-{scenario_ix}" if scenario.scenario_id is None else scenario.scenario_id
        used = ids_used.get(scenario_id) is not None
        if used:
            raise ValueError(f"Duplicate scenario ID found: {scenario_id}")

        ids_used[scenario_id] = True
        scenario_by_id[scenario_id] = scenario

    return scenario_by_id



import nextmv

options = nextmv.Options(
    nextmv.Option("duration", str, "30s", description="solver duration"),
    nextmv.Option("threads", int, 4, description="computer threads"),
)

print(options.to_dict())



"""
Module for handling output destinations and data.

This module provides classes and functions for handling the output of decision
problems, including formatting, serialization, and writing to various
destinations.

Classes
-------
RunStatistics
    Statistics about a general run.
ResultStatistics
    Statistics about a specific result.
DataPoint
    A data point representing a 2D coordinate.
Series
    A series of data points for visualization or analysis.
SeriesData
    Data container for multiple series of data points.
Statistics
    Complete statistics container for a solution, including run metrics and result data.
OutputFormat
    Enumeration of supported output formats.
SolutionFile
    Represents a solution to be written as a file.
VisualSchema
    Enumeration of supported visualization schemas.
Visual
    Visual schema definition for an asset.
Asset
    Represents downloadable information that is part of the `Output`.
Output
    A class for representing the output of a decision problem.
OutputWriter
    Base class for writing outputs to different destinations.
LocalOutputWriter
    Class for writing outputs to local files or stdout.

Functions
---------
write
    Write the output to the specified destination.

Attributes
----------
ASSETS_KEY : str
    Assets key constant used for identifying assets in the run output.
STATISTICS_KEY : str
    Statistics key constant used for identifying statistics in the run output.
SOLUTIONS_KEY : str
    Solutions key constant used for identifying solutions in the run output.
OUTPUTS_KEY : str
    Outputs key constant used for identifying outputs in the run output.
"""

import copy
import csv
import os
import sys
from collections.abc import Callable
from dataclasses import dataclass
from enum import Enum
from typing import Any

from pydantic import AliasChoices, Field

from nextmv._serialization import serialize_json
from nextmv.base_model import BaseModel
from nextmv.deprecated import deprecated
from nextmv.logger import reset_stdout
from nextmv.options import Options

ASSETS_KEY = "assets"
"""
Assets key constant used for identifying assets in the run output.
"""
STATISTICS_KEY = "statistics"
"""
Statistics key constant used for identifying statistics in the run output.
"""
SOLUTIONS_KEY = "solutions"
"""
Solutions key constant used for identifying solutions in the run output.
"""
OUTPUTS_KEY = "outputs"
"""
Outputs key constant used for identifying outputs in the run output.
"""


class RunStatistics(BaseModel):
    """
    Statistics about a general run.

    You can import the `RunStatistics` class directly from `nextmv`:

    ```python
    from nextmv import RunStatistics
    ```

    Parameters
    ----------
    duration : float, optional
        Duration of the run in seconds.
    iterations : int, optional
        Number of iterations.
    custom : Union[Any, dict[str, Any]], optional
        Custom statistics created by the user. Can normally expect a `dict[str,
        Any]`.

    Examples
    --------
    >>> from nextmv.output import RunStatistics
    >>> stats = RunStatistics(duration=10.5, iterations=100)
    >>> stats.duration
    10.5
    >>> stats.custom = {"convergence": 0.001}
    >>> stats.to_dict()
    {'duration': 10.5, 'iterations': 100, 'custom': {'convergence': 0.001}}
    """

    duration: float | None = None
    """Duration of the run in seconds."""
    iterations: int | None = None
    """Number of iterations."""
    custom: Any | dict[str, Any] | None = None
    """Custom statistics created by the user. Can normally expect a `dict[str,
    Any]`."""


class ResultStatistics(BaseModel):
    """
    Statistics about a specific result.

    You can import the `ResultStatistics` class directly from `nextmv`:

    ```python
    from nextmv import ResultStatistics
    ```

    Parameters
    ----------
    duration : float, optional
        Duration of the run in seconds.
    value : float, optional
        Value of the result.
    custom : Union[Any, dict[str, Any]], optional
        Custom statistics created by the user. Can normally expect a `dict[str,
        Any]`.

    Examples
    --------
    >>> from nextmv.output import ResultStatistics
    >>> result_stats = ResultStatistics(duration=5.2, value=42.0)
    >>> result_stats.value
    42.0
    >>> result_stats.custom = {"gap": 0.05}
    >>> result_stats.to_dict()
    {'duration': 5.2, 'value': 42.0, 'custom': {'gap': 0.05}}
    """

    duration: float | None = None
    """Duration of the run in seconds."""
    value: float | None = None
    """Value of the result."""
    custom: Any | dict[str, Any] | None = None
    """Custom statistics created by the user. Can normally expect a `dict[str,
    Any]`."""


class DataPoint(BaseModel):
    """
    A data point representing a 2D coordinate.

    You can import the `DataPoint` class directly from `nextmv`:

    ```python
    from nextmv import DataPoint
    ```

    Parameters
    ----------
    x : float
        X coordinate of the data point.
    y : float
        Y coordinate of the data point.

    Examples
    --------
    >>> from nextmv.output import DataPoint
    >>> point = DataPoint(x=3.5, y=4.2)
    >>> point.x
    3.5
    >>> point.to_dict()
    {'x': 3.5, 'y': 4.2}
    """

    x: float
    """X coordinate of the data point."""
    y: float
    """Y coordinate of the data point."""


class Series(BaseModel):
    """
    A series of data points for visualization or analysis.

    You can import the `Series` class directly from `nextmv`:

    ```python
    from nextmv import Series
    ```

    Parameters
    ----------
    name : str, optional
        Name of the series.
    data_points : list[DataPoint], optional
        Data points of the series.

    Examples
    --------
    >>> from nextmv.output import Series, DataPoint
    >>> points = [DataPoint(x=1.0, y=2.0), DataPoint(x=2.0, y=3.0)]
    >>> series = Series(name="Example Series", data_points=points)
    >>> series.name
    'Example Series'
    >>> len(series.data_points)
    2
    """

    name: str | None = None
    """Name of the series."""
    data_points: list[DataPoint] | None = None
    """Data of the series."""


class SeriesData(BaseModel):
    """
    Data container for multiple series of data points.

    You can import the `SeriesData` class directly from `nextmv`:

    ```python
    from nextmv import SeriesData
    ```

    Parameters
    ----------
    value : Series, optional
        A series for the value of the solution.
    custom : list[Series], optional
        A list of series for custom statistics.

    Examples
    --------
    >>> from nextmv.output import SeriesData, Series, DataPoint
    >>> value_series = Series(name="Solution Value", data_points=[DataPoint(x=0, y=10), DataPoint(x=1, y=5)])
    >>> custom_series = [Series(name="Gap", data_points=[DataPoint(x=0, y=0.5), DataPoint(x=1, y=0.1)])]
    >>> series_data = SeriesData(value=value_series, custom=custom_series)
    >>> series_data.value.name
    'Solution Value'
    >>> len(series_data.custom)
    1
    """

    value: Series | None = None
    """A series for the value of the solution."""
    custom: list[Series] | None = None
    """A list of series for custom statistics."""


class Statistics(BaseModel):
    """
    Complete statistics container for a solution, including run metrics and
    result data.

    You can import the `Statistics` class directly from `nextmv`:

    ```python
    from nextmv import Statistics
    ```

    Parameters
    ----------
    run : RunStatistics, optional
        Statistics about the run.
    result : ResultStatistics, optional
        Statistics about the last result.
    series_data : SeriesData, optional
        Series data about some metric.
    statistics_schema : str, optional
        Schema (version). This class only supports `v1`.

    Examples
    --------
    >>> from nextmv.output import Statistics, RunStatistics, ResultStatistics
    >>> run_stats = RunStatistics(duration=10.0, iterations=50)
    >>> result_stats = ResultStatistics(value=100.0)
    >>> stats = Statistics(run=run_stats, result=result_stats, statistics_schema="v1")
    >>> stats.run.duration
    10.0
    >>> stats.result.value
    100.0
    """

    run: RunStatistics | None = None
    """Statistics about the run."""
    result: ResultStatistics | None = None
    """Statistics about the last result."""
    series_data: SeriesData | None = None
    """Data of the series."""
    statistics_schema: str | None = Field(
        serialization_alias="schema",
        validation_alias=AliasChoices("schema", "statistics_schema"),
        default="v1",
    )
    """Schema (version). This class only supports `v1`."""


class VisualSchema(str, Enum):
    """
    Enumeration of supported visualization schemas.

    You can import the `VisualSchema` class directly from `nextmv`:

    ```python
    from nextmv import VisualSchema
    ```

    This enum defines the different visualization libraries or rendering methods
    that can be used to display custom asset data in the Nextmv Console.

    Attributes
    ----------
    CHARTJS : str
        Tells Nextmv Console to render the custom asset data with the Chart.js library.
    GEOJSON : str
        Tells Nextmv Console to render the custom asset data as GeoJSON on a map.
    PLOTLY : str
        Tells Nextmv Console to render the custom asset data with the Plotly library.
    """

    CHARTJS = "chartjs"
    """Tells Nextmv Console to render the custom asset data with the Chart.js
    library."""
    GEOJSON = "geojson"
    """Tells Nextmv Console to render the custom asset data as GeoJSON on a
    map."""
    PLOTLY = "plotly"
    """Tells Nextmv Console to render the custom asset data with the Plotly
    library."""


class Visual(BaseModel):
    """
    Visual schema definition for an asset.

    You can import the `Visual` class directly from `nextmv`:

    ```python
    from nextmv import Visual
    ```

    This class defines how an asset is plotted in the Nextmv Console,
    including the schema type, label, and display type.

    Parameters
    ----------
    visual_schema : VisualSchema
        Schema of the visual asset.
    label : str
        Label for the custom tab of the visual asset in the Nextmv Console.
    visual_type : str, optional
        Defines the type of custom visual. Default is "custom-tab".

    Raises
    ------
    ValueError
        If an unsupported schema or visual_type is provided.

    Examples
    --------
    >>> from nextmv.output import Visual, VisualSchema
    >>> visual = Visual(visual_schema=VisualSchema.CHARTJS, label="Performance Chart")
    >>> visual.visual_schema
    <VisualSchema.CHARTJS: 'chartjs'>
    >>> visual.label
    'Performance Chart'
    """

    visual_schema: VisualSchema = Field(
        serialization_alias="schema",
        validation_alias=AliasChoices("schema", "visual_schema"),
    )
    """Schema of the visual asset."""
    label: str
    """Label for the custom tab of the visual asset in the Nextmv Console."""

    visual_type: str | None = Field(
        serialization_alias="type",
        validation_alias=AliasChoices("type", "visual_type"),
        default="custom-tab",
    )
    """Defines the type of custom visual, currently there is only one type:
    `custom-tab`. This renders the visual in its own tab view of the run
    details."""

    def __post_init__(self):
        """
        Validate the visual schema and type.

        Raises
        ------
        ValueError
            If the visual_schema is not in VisualSchema or if visual_type is not 'custom-tab'.
        """
        if self.visual_schema not in VisualSchema:
            raise ValueError(f"unsupported schema: {self.visual_schema}, supported schemas are {VisualSchema}")

        if self.visual_type != "custom-tab":
            raise ValueError(f"unsupported visual_type: {self.visual_type}, supported types are `custom-tab`")


class Asset(BaseModel):
    """
    Represents downloadable information that is part of the `Output`.

    You can import the `Asset` class directly from `nextmv`:

    ```python
    from nextmv import Asset
    ```

    An asset contains content that can be serialized to JSON and optionally
    includes visual information for rendering in the Nextmv Console.

    Parameters
    ----------
    name : str
        Name of the asset.
    content : Any
        Content of the asset. The type must be serializable to JSON.
    content_type : str, optional
        Content type of the asset. Only "json" is currently supported. Default is "json".
    description : str, optional
        Description of the asset. Default is None.
    visual : Visual, optional
        Visual schema of the asset. Default is None.

    Raises
    ------
    ValueError
        If the content_type is not "json".

    Examples
    --------
    >>> from nextmv.output import Asset, Visual, VisualSchema
    >>> visual = Visual(visual_schema=VisualSchema.CHARTJS, label="Solution Progress")
    >>> asset = Asset(
    ...     name="optimization_progress",
    ...     content={"iterations": [1, 2, 3], "values": [10, 8, 7]},
    ...     description="Optimization progress over iterations",
    ...     visual=visual
    ... )
    >>> asset.name
    'optimization_progress'
    """

    name: str
    """Name of the asset."""

    id: str | None = None
    """
    The ID of the asset. This ID will be populated by the Nextmv platform and can be used
    to download the asset later.
    """
    content: Any | None = None
    """
    Content of the asset. The type must be serializable to JSON. Can be empty when
    fetching the asset metadata only (e.g.: via the asset list endpoint).
    """
    content_type: str | None = "json"
    """Content type of the asset. Only `json` is allowed"""
    description: str | None = None
    """Description of the asset."""
    visual: Visual | None = None
    """Visual schema of the asset."""

    def __post_init__(self):
        """
        Validate the content type.

        Raises
        ------
        ValueError
            If the content_type is not "json".
        """
        if self.content_type != "json":
            raise ValueError(f"unsupported content_type: {self.content_type}, supported types are `json`")


class OutputFormat(str, Enum):
    """
    Enumeration of supported output formats.

    You can import the `OutputFormat` class directly from `nextmv`:

    ```python
    from nextmv import OutputFormat
    ```

    This enum defines the different formats that can be used for outputting data.
    Each format has specific requirements and behaviors when writing.

    Attributes
    ----------
    JSON : str
        JSON format, utf-8 encoded.
    CSV_ARCHIVE : str
        CSV archive format: multiple CSV files.
    MULTI_FILE : str
        Multi-file format: multiple files in a directory.
    TEXT : str
        Text format, utf-8 encoded.
    """

    JSON = "json"
    """JSON format, utf-8 encoded."""
    CSV_ARCHIVE = "csv-archive"
    """CSV archive format: multiple CSV files."""
    MULTI_FILE = "multi-file"
    """Multi-file format: multiple files in a directory."""
    TEXT = "text"
    """Text format, utf-8 encoded."""


@dataclass
class SolutionFile:
    """
    Represents a solution to be written as a file.

    You can import the `SolutionFile` class directly from `nextmv`:

    ```python
    from nextmv import SolutionFile
    ```

    This class is used to define a solution that will be written to a file in
    the filesystem. It includes the name of the file, the data to be written,
    and the writer function that will handle the serialization of the data.
    This `SolutionFile` class is typically used in the `Output`, when the
    `Output.output_format` is set to `OutputFormat.MULTI_FILE`. Given that it
    is difficult to handle every edge case of how a solution is serialized, and
    written to a file, this class exists so that the user can implement the
    `writer` callable of their choice and provide it with any `writer_args`
    and `writer_kwargs` they might need.

    Parameters
    ----------
    name : str
        Name of the output file. The file extension should be included in the
        name.
    data : Any
        The actual data that will be written to the file. This can be any type
        that can be given to the `writer` function. For example, if the `writer`
        is a `csv.DictWriter`, then the data should be a list of dictionaries,
        where each dictionary represents a row in the CSV file.
    writer : Callable
        Callable that writes the solution data to the file. This should be a
        function implemented by the user. There are convenience functions that you
        can use as a writer as well. The `writer` must receive, at the very
        minimum, the following arguments:

        - `file_path`: a `str` argument which is the location where this solution
        will be written to. This includes the dir and the name of the file. As
        such, the `name` parameter of this class is going to be passed to this
        function joined with the directory where the file will be written.
        - `data`: the actual data that will be written to the file. This can be any
        type that can be given to the `writer` function. The `data` parameter of
        this class is going to be passed to the `writer` function.

        The `writer` can also receive additional arguments, and keyword arguments.
        The `writer_args` and `writer_kwargs` parameters of this class can be used
        to provide those additional arguments.
    writer_args : Optional[list[Any]], optional
        Positional arguments to pass to the writer function.
    writer_kwargs : Optional[dict[str, Any]], optional
        Keyword arguments to pass to the writer function.

    Examples
    --------
    >>> from nextmv import SolutionFile
    >>> solution_file = SolutionFile(
    ...     name="solution.csv",
    ...     data=[{"id": 1, "value": 100}, {"id": 2, "value": 200}],
    ...     writer=csv.DictWriter,
    ...     writer_kwargs={"fieldnames": ["id", "value"]},
    ...     writer_args=[open("solution.csv", "w", newline="")],
    ... )
    """

    name: str
    """
    Name of the solution (output) file. The file extension should be included in the
    name.
    """
    data: Any
    """
    The actual data that will be written to the file. This can be any type that
    can be given to the `writer` function. For example, if the `writer` is a
    `csv.DictWriter`, then the data should be a list of dictionaries, where
    each dictionary represents a row in the CSV file.
    """
    writer: Callable[[str, Any], None]
    """
    Callable that writes the solution data to the file. This should be a
    function implemented by the user. There are convenience functions that you
    can use as a writer as well. The `writer` must receive, at the very
    minimum, the following arguments:

    - `file_path`: a `str` argument which is the location where this solution
      will be written to. This includes the dir and the name of the file. As
      such, the `name` parameter of this class is going to be passed to this
      function joined with the directory where the file will be written.
    - `data`: the actual data that will be written to the file. This can be any
      type that can be given to the `writer` function. The `data` parameter of
      this class is going to be passed to the `writer` function.

    The `writer` can also receive additional arguments, and keyword arguments.
    The `writer_args` and `writer_kwargs` parameters of this class can be used
    to provide those additional arguments.
    """
    writer_args: list[Any] | None = None
    """
    Optional positional arguments to pass to the writer function. This can be
    used to customize the behavior of the writer.
    """
    writer_kwargs: dict[str, Any] | None = None
    """
    Optional keyword arguments to pass to the writer function. This can be used
    to customize the behavior of the writer.
    """


def json_solution_file(
    name: str,
    data: dict[str, Any],
    json_configurations: dict[str, Any] | None = None,
) -> SolutionFile:
    """
    This is a convenience function to build a `SolutionFile`. It writes the
    given `data` to a `.json` file with the provided `name`.

    You can import this function directly from `nextmv`:

    ```python
    from nextmv import json_solution_file
    ```

    Parameters
    ----------
    name : str
        Name of the output file. You don't need to include the `.json`
        extension.
    data : dict[str, Any]
        The actual data that will be written to the file. This should be a
        dictionary that can be serialized to JSON.
    json_configurations : Optional[dict[str, Any]], optional
        Optional configuration options for the JSON serialization process. You
        can use these options to configure parameters such as indentation.

    Returns
    -------
    SolutionFile
        The constructed `SolutionFile` object.

    Examples
    --------
    >>> from nextmv import json_solution_file
    >>> solution_file = json_solution_file(
    ...     name="solution",
    ...     data={"id": 1, "value": 100}
    ... )
    >>> solution_file.name
    'solution.json'
    >>> solution_file.data
    {'id': 1, 'value': 100}
    """

    if not name.endswith(".json"):
        name += ".json"

    json_configurations = json_configurations or {}

    def writer(file_path: str, write_data: dict[str, Any]) -> None:
        serialized = serialize_json(write_data, json_configurations=json_configurations)

        with open(file_path, "w", encoding="utf-8") as file:
            file.write(serialized + "\n")

    return SolutionFile(
        name=name,
        data=data,
        writer=writer,
    )


def csv_solution_file(
    name: str,
    data: list[dict[str, Any]],
    csv_configurations: dict[str, Any] | None = None,
) -> SolutionFile:
    """
    This is a convenience function to build a `SolutionFile`. It writes the
    given `data` to a `.csv` file with the provided `name`.

    You can import this function directly from `nextmv`:

    ```python
    from nextmv import csv_solution_file
    ```

    Parameters
    ----------
    name : str
        Name of the output file. You don't need to include the `.csv`
        extension.
    data : list[dict[str, Any]]
        The actual data that will be written to the file. This should be a list
        of dictionaries, where each dictionary represents a row in the CSV file.
        The keys of the dictionaries will be used as the column headers in the
        CSV file.
    csv_configurations : Optional[dict[str, Any]], optional
        Optional configuration options for the CSV serialization process.

    Returns
    -------
    SolutionFile
        The constructed `SolutionFile` object.

    Examples
    --------
    >>> from nextmv import csv_solution_file
    >>> solution_file = csv_solution_file(
    ...     name="solution",
    ...     data=[{"id": 1, "value": 100}, {"id": 2, "value": 200}]
    ... )
    >>> solution_file.name
    'solution.csv'
    >>> solution_file.data
    [{'id': 1, 'value': 100}, {'id': 2, 'value': 200}]
    """

    if not name.endswith(".csv"):
        name += ".csv"

    csv_configurations = csv_configurations or {}

    def writer(file_path: str, write_data: list[dict[str, Any]]) -> None:
        with open(file_path, "w", encoding="utf-8", newline="") as file:
            writer = csv.DictWriter(
                file,
                fieldnames=write_data[0].keys(),
                **csv_configurations,
            )
            writer.writeheader()
            writer.writerows(write_data)

    return SolutionFile(
        name=name,
        data=data,
        writer=writer,
    )


def text_solution_file(name: str, data: str) -> SolutionFile:
    """
    This is a convenience function to build a `SolutionFile`. It writes the
    given `data` to a utf-8 encoded file with the provided `name`.

    You can import this function directly from `nextmv`:

    ```python
    from nextmv import text_solution_file
    ```

    You must provide the extension as part of the `name` parameter.

    Parameters
    ----------
    name : str
        Name of the output file. The file extension must be provided in the
        name.
    data : str
        The actual data that will be written to the file.

    Returns
    -------
    SolutionFile
        The constructed `SolutionFile` object.

    Examples
    --------
    >>> from nextmv import text_solution_file
    >>> solution_file = text_solution_file(
    ...     name="solution.txt",
    ...     data="This is a sample text solution."
    ... )
    >>> solution_file.name
    'solution.txt'
    >>> solution_file.data
    'This is a sample text solution.'
    """

    def writer(file_path: str, write_data: str) -> None:
        with open(file_path, "w", encoding="utf-8") as file:
            file.write(write_data + "\n")

    return SolutionFile(
        name=name,
        data=data,
        writer=writer,
    )


@dataclass
class Output:
    """
    Output of a decision problem.

    You can import the `Output` class directly from `nextmv`:

    ```python
    from nextmv import Output
    ```

    This class is used to structure the output of a decision problem that
    can later be written to various destinations. It supports different output
    formats and allows for customization of the serialization process.

    The `solution`'s type must match the `output_format`:

    - `OutputFormat.JSON`: the data must be `dict[str, Any]` or `Any`.
    - `OutputFormat.CSV_ARCHIVE`: the data must be `dict[str, list[dict[str,
       Any]]]`. The keys represent the file names where the data should be
       written. The values are lists of dictionaries, where each dictionary
       represents a row in the CSV file.

    If you are working with `OutputFormat.MULTI_FILE`, you should use
    `solution_files` instead of `solution`. When `solution_files` is not
    `None`, then the `output_format` _must_ be `OutputFormat.MULTI_FILE`.
    `solution_files` is a list of `SolutionFile` objects, which allows you to
    define the name of the file, the data to be written, and the writer
    function that will handle the serialization of the data. This is useful when
    you need to write the solution to multiple files with different formats or
    configurations.

    There are convenience functions to create `SolutionFile` objects for
    common use cases, such as:

    - `json_solution_file`: for writing JSON data to a file.
    - `csv_solution_file`: for writing CSV data to a file.
    - `text_solution_file`: for writing utf-8 encoded data to a file.

    For other data types, such as Excel, you can create your own `SolutionFile`
    objects by providing a `name`, `data`, and a `writer` function that will
    handle the serialization of the data.

    Parameters
    ----------
    options : Optional[Union[Options, dict[str, Any]]], optional
        Options that the `Output` was created with. These options can be of type
        `Options` or a simple dictionary. Default is None.
    output_format : Optional[OutputFormat], optional
        Format of the output data. Default is `OutputFormat.JSON`.
    solution : Optional[Union[dict[str, Any], Any, dict[str, list[dict[str, Any]]]]], optional
        The solution to the decision problem. The type must match the
        `output_format`. Default is None.
    statistics : Optional[Union[Statistics, dict[str, Any]]], optional
        Statistics of the solution. Default is None.
    csv_configurations : Optional[dict[str, Any]], optional
        Configuration for writing CSV files. Default is None.
    json_configurations : Optional[dict[str, Any]], optional
        Configuration for writing JSON files. Default is None.
    assets : Optional[list[Union[Asset, dict[str, Any]]]], optional
        List of assets to be included in the output. Default is None.
    solution_files: Optional[list[SolutionFile]], default = None
        Optional list of solution files to be included in the output. These
        files are of type `SolutionFile`, which allows for custom serialization
        and writing of the solution data to files. When this field is
        specified, then the `output_format` must be set to
        `OutputFormat.MULTI_FILE`, otherwise an exception will be raised. The
        `SolutionFile` class allows you to define the name of the file, the
        data to be written, and the writer function that will handle the
        serialization of the data. This is useful when you need to write the
        solution to multiple files with different formats or configurations.

        There are convenience functions to create `SolutionFile` objects for
        common use cases, such as:

        - `json_solution_file`: for writing JSON data to a file.
        - `csv_solution_file`: for writing CSV data to a file.
        - `text_solution_file`: for writing utf-8 encoded data to a file.

        For other data types, such as Excel, you can create your own
        `SolutionFile` objects by providing a `name`, `data`, and a `writer`
        function that will handle the serialization of the data.

    Raises
    ------
    ValueError
        If the solution is not compatible with the specified output_format.
    TypeError
        If options, statistics, or assets have unsupported types.

    Examples
    --------
    >>> from nextmv.output import Output, OutputFormat, Statistics, RunStatistics
    >>> run_stats = RunStatistics(duration=30.0, iterations=100)
    >>> stats = Statistics(run=run_stats)
    >>> solution = {"routes": [{"vehicle": 1, "stops": [1, 2, 3]}, {"vehicle": 2, "stops": [4, 5]}]}
    >>> output = Output(
    ...     output_format=OutputFormat.JSON,
    ...     solution=solution,
    ...     statistics=stats,
    ...     json_configurations={"indent": 4}
    ... )
    >>> output_dict = output.to_dict()
    >>> "solution" in output_dict and "statistics" in output_dict
    True
    """

    options: Options | dict[str, Any] | None = None
    """
    Options that the `Output` was created with. These options can be of type
    `Options` or a simple dictionary. If the options are of type `Options`,
    they will be serialized to a dictionary using the `to_dict` method. If
    they are a dictionary, they will be used as is. If the options are not
    provided, an empty dictionary will be used. If the options are of type
    `dict`, then the dictionary should have the following structure:

    ```python
    {
        "duration": "30",
        "threads": 4,
    }
    ```
    """
    output_format: OutputFormat | None = OutputFormat.JSON
    """
    Format of the output data. Default is `OutputFormat.JSON`. When set to
    `OutputFormat.MULTI_FILE`, the `solution_files` field must be specified and
    cannot be `None`.
    """
    solution: dict[str, Any] | Any | dict[str, list[dict[str, Any]]] | None = None
    """
    The solution to the decision problem. Use this filed when working with
    `output_format` of types:

    - `OutputFormat.JSON`: the data must be `dict[str, Any]` or `Any`.
    - `OutputFormat.CSV_ARCHIVE`: the data must be `dict[str, list[dict[str,
    Any]]]`. The keys represent the file names where the data will be written
    to. The values are lists of dictionaries, where each dictionary represents
    a row in the CSV file.

    Note that when the `output_format` is set to `OutputFormat.MULTI_FILE`,
    this `solution` field is ignored, as you should use the `solution_files`
    field instead.
    """
    statistics: Statistics | dict[str, Any] | None = None
    """
    Statistics of the solution. These statistics can be of type `Statistics` or a
    simple dictionary. If the statistics are of type `Statistics`, they will be
    serialized to a dictionary using the `to_dict` method. If they are a
    dictionary, they will be used as is. If the statistics are not provided, an
    empty dictionary will be used.
    """
    csv_configurations: dict[str, Any] | None = None
    """
    Optional configuration for writing CSV files, to be used when the
    `output_format` is `OutputFormat.CSV_ARCHIVE`. These configurations are
    passed as kwargs to the `DictWriter` class from the `csv` module.
    """
    json_configurations: dict[str, Any] | None = None
    """
    Optional configuration for writing JSON files, to be used when the
    `output_format` is `OutputFormat.JSON`. These configurations are passed as
    kwargs to the `json.dumps` function.
    """
    assets: list[Asset | dict[str, Any]] | None = None
    """
    Optional list of assets to be included in the output. These assets can be of
    type `Asset` or a simple dictionary. If the assets are of type `Asset`, they
    will be serialized to a dictionary using the `to_dict` method. If they are a
    dictionary, they will be used as is. If the assets are not provided, an
    empty list will be used.
    """
    solution_files: list[SolutionFile] | None = None
    """
    Optional list of solution files to be included in the output. These files
    are of type `SolutionFile`, which allows for custom serialization and
    writing of the solution data to files. When this field is specified, then
    the `output_format` must be set to `OutputFormat.MULTI_FILE`, otherwise an
    exception will be raised. The `SolutionFile` class allows you to define the
    name of the file, the data to be written, and the writer function that will
    handle the serialization of the data. This is useful when you need to write
    the solution to multiple files with different formats or configurations.

    There are convenience functions to create `SolutionFile` objects for
    common use cases, such as:

    - `json_solution_file`: for writing JSON data to a file.
    - `csv_solution_file`: for writing CSV data to a file.
    - `text_solution_file`: for writing utf-8 encoded data to a file.

    For other data types, such as Excel, you can create your own `SolutionFile`
    objects by providing a `name`, `data`, and a `writer` function that will
    handle the serialization of the data.
    """

    def __post_init__(self):
        """
        Initialize and validate the Output instance.

        This method performs two main tasks:
        1. Creates a deep copy of the options to preserve the original values
        2. Validates that the solution matches the specified output_format

        Raises
        ------
        ValueError
            If the solution is not compatible with the specified output_format.
        """
        # Capture a snapshot of the options that were used to create the class
        # so even if they are changed later, we have a record of the original.
        init_options = self.options
        new_options = copy.deepcopy(init_options)
        self.options = new_options

        if self.solution is not None:
            if self.output_format == OutputFormat.JSON:
                try:
                    _ = serialize_json(self.solution)
                except (TypeError, OverflowError) as e:
                    raise ValueError(
                        f"Output has output_format OutputFormat.JSON and "
                        f"Output.solution is of type {type(self.solution)}, which is not JSON serializable"
                    ) from e

            elif self.output_format == OutputFormat.CSV_ARCHIVE and not isinstance(self.solution, dict):
                raise ValueError(
                    f"unsupported Output.solution type: {type(self.solution)} with "
                    "output_format OutputFormat.CSV_ARCHIVE, supported type is `dict`"
                )

        if self.solution_files is not None and self.output_format != OutputFormat.MULTI_FILE:
            raise ValueError(
                f"`solution_files` are not `None`, but `output_format` is different from `OutputFormat.MULTI_FILE`: "
                f"{self.output_format}. If you want to use `solution_files`, set `output_format` "
                "to `OutputFormat.MULTI_FILE`."
            )
        elif self.solution_files is not None and not isinstance(self.solution_files, list):
            raise TypeError(
                f"unsupported Output.solution_files type: {type(self.solution_files)}, supported type is `list`"
            )

    def to_dict(self) -> dict[str, Any]:  # noqa: C901
        """
        Convert the `Output` object to a dictionary.

        Returns
        -------
        dict[str, Any]
            The dictionary representation of the `Output` object.
        """

        # Options need to end up as a dict, so we achieve that based on the
        # type of options that were used to create the class.
        if self.options is None:
            options = {}
        elif isinstance(self.options, Options):
            options = self.options.to_dict()
        elif isinstance(self.options, dict):
            options = self.options
        else:
            raise TypeError(f"unsupported options type: {type(self.options)}, supported types are `Options` or `dict`")

        # Statistics need to end up as a dict, so we achieve that based on the
        # type of statistics that were used to create the class.
        if self.statistics is None:
            statistics = {}
        elif isinstance(self.statistics, Statistics):
            statistics = self.statistics.to_dict()
        elif isinstance(self.statistics, dict):
            statistics = self.statistics
        else:
            raise TypeError(
                f"unsupported statistics type: {type(self.statistics)}, supported types are `Statistics` or `dict`"
            )

        # Assets need to end up as a list of dicts, so we achieve that based on
        # the type of each asset in the list.
        assets = []
        if isinstance(self.assets, list):
            for ix, asset in enumerate(self.assets):
                if isinstance(asset, Asset):
                    assets.append(asset.to_dict())
                elif isinstance(asset, dict):
                    assets.append(asset)
                else:
                    raise TypeError(
                        f"unsupported asset {ix}, type: {type(asset)}; supported types are `Asset` or `dict`"
                    )
        elif self.assets is not None:
            raise TypeError(f"unsupported assets type: {type(self.assets)}, supported types are `list`")

        output_dict = {
            "options": options,
            "solution": self.solution if self.solution is not None else {},
            STATISTICS_KEY: statistics,
            ASSETS_KEY: assets,
        }

        # Add the auxiliary configurations to the output dictionary if they are
        # defined and not empty.
        if (
            self.output_format == OutputFormat.CSV_ARCHIVE
            and self.csv_configurations is not None
            and self.csv_configurations != {}
        ):
            output_dict["csv_configurations"] = self.csv_configurations

        return output_dict


class OutputWriter:
    """
    Base class for writing outputs.

    You can import the `OutputWriter` class directly from `nextmv`:

    ```python
    from nextmv import OutputWriter
    ```

    This is an abstract base class that defines the interface for writing outputs
    to different destinations. Subclasses should implement the `write` method.

    Examples
    --------
    >>> class CustomOutputWriter(OutputWriter):
    ...     def write(self, output, path=None, **kwargs):
    ...         # Custom implementation for writing output
    ...         print(f"Writing output to {path}")
    """

    def write(self, output: Output | dict[str, Any] | BaseModel, *args, **kwargs) -> None:
        """
        Write the output data.

        This is an abstract method that should be implemented by subclasses.

        Parameters
        ----------
        output : Union[Output, dict[str, Any], BaseModel]
            The output data to write.
        *args
            Variable length argument list.
        **kwargs
            Arbitrary keyword arguments.

        Raises
        ------
        NotImplementedError
            This method must be implemented by subclasses.
        """
        raise NotImplementedError


class LocalOutputWriter(OutputWriter):
    """
    Class for writing outputs to local files or stdout.

    You can import the `LocalOutputWriter` class directly from `nextmv`:

    ```python
    from nextmv import LocalOutputWriter
    ```

    This class implements the OutputWriter interface to write output data to
    local files or stdout. The destination and format depend on the output
    format and the provided path.

    Examples
    --------
    >>> from nextmv.output import LocalOutputWriter, Output, Statistics
    >>> writer = LocalOutputWriter()
    >>> output = Output(solution={"result": 42}, statistics=Statistics())
    >>> # Write to stdout
    >>> writer.write(output, path=None)
    >>> # Write to a file
    >>> writer.write(output, path="results.json")
    """

    def _write_json(
        self,
        output: Output | dict[str, Any] | BaseModel,
        output_dict: dict[str, Any],
        path: str | None = None,
    ) -> None:
        """
        Write output in JSON format.

        Parameters
        ----------
        output : Union[Output, dict[str, Any], BaseModel]
            The output object containing configuration.
        output_dict : dict[str, Any]
            Dictionary representation of the output to write.
        path : str, optional
            Path to write the output. If None or empty, writes to stdout.
        """
        json_configurations = {}
        if hasattr(output, "json_configurations") and output.json_configurations is not None:
            json_configurations = output.json_configurations

        serialized = serialize_json(
            output_dict,
            json_configurations=json_configurations,
        )

        if path is None or path == "":
            print(serialized, file=sys.stdout)
            return

        with open(path, "w", encoding="utf-8") as file:
            file.write(serialized + "\n")

    def _write_archive(
        self,
        output: Output | dict[str, Any] | BaseModel,
        output_dict: dict[str, Any],
        path: str | None = None,
    ) -> None:
        """
        Write output in CSV archive format.

        Parameters
        ----------
        output : Union[Output, dict[str, Any], BaseModel]
            The output object containing configuration and solution data.
        output_dict : dict[str, Any]
            Dictionary representation of the output to write.
        path : str, optional
            Directory path to write the CSV files. If None or empty,
            writes to a directory named "output" in the current working directory.

        Raises
        ------
        ValueError
            If the path is an existing file instead of a directory.
        """
        dir_path = "output"
        if path is not None and path != "":
            if os.path.isfile(path):
                raise ValueError(f"The path refers to an existing file: {path}")

            dir_path = path

        if not os.path.exists(dir_path):
            os.makedirs(dir_path)

        json_configurations = {}
        if hasattr(output, "json_configurations") and output.json_configurations is not None:
            json_configurations = output.json_configurations

        serialized = serialize_json(
            {
                "options": output_dict.get("options", {}),
                STATISTICS_KEY: output_dict.get(STATISTICS_KEY, {}),
                ASSETS_KEY: output_dict.get(ASSETS_KEY, []),
            },
            json_configurations=json_configurations,
        )
        print(serialized, file=sys.stdout)

        if output.solution is None:
            return

        csv_configurations = output.csv_configurations
        if csv_configurations is None:
            csv_configurations = {}

        for file_name, data in output.solution.items():
            file_path = os.path.join(dir_path, f"{file_name}.csv")
            with open(file_path, "w", encoding="utf-8", newline="") as file:
                writer = csv.DictWriter(
                    file,
                    fieldnames=data[0].keys(),
                    **csv_configurations,
                )
                writer.writeheader()
                writer.writerows(data)

    def _write_multi_file(
        self,
        output: Output | dict[str, Any] | BaseModel,
        output_dict: dict[str, Any],
        path: str | None = None,
    ) -> None:
        """
        Write output to multiple files.

        Parameters
        ----------
        output : Union[Output, dict[str, Any], BaseModel]
            The output object containing configuration and solution data.
        output_dict : dict[str, Any]
            Dictionary representation of the output to write.
        path : str, optional
            Directory path to write the CSV files. If None or empty,
            writes to a directory named "output" in the current working directory.

        Raises
        ------
        ValueError
            If the path is an existing file instead of a directory.
        """
        dir_path = OUTPUTS_KEY
        if path is not None and path != "":
            if os.path.isfile(path):
                raise ValueError(f"The path refers to an existing file: {path}")

            dir_path = path

        if not os.path.exists(dir_path):
            os.makedirs(dir_path)

        json_configurations = {}
        if hasattr(output, "json_configurations") and output.json_configurations is not None:
            json_configurations = output.json_configurations

        self._write_multi_file_element(
            parent_dir=dir_path,
            json_configurations=json_configurations,
            output_dict=output_dict,
            element_key=STATISTICS_KEY,
        )
        self._write_multi_file_element(
            parent_dir=dir_path,
            json_configurations=json_configurations,
            output_dict=output_dict,
            element_key=ASSETS_KEY,
        )
        self._write_multi_file_solution(dir_path=dir_path, output=output)

    def _write_multi_file_element(
        self,
        parent_dir: str,
        output_dict: dict[str, Any],
        element_key: str,
        json_configurations: dict[str, Any] | None = None,
    ):
        """
        Auxiliary function to write a specific element of the output
        dictionary to a file in the specified parent directory.
        """

        element = output_dict.get(element_key)
        if element is None or not element:
            return

        final_dir = os.path.join(parent_dir, element_key)

        if not os.path.exists(final_dir):
            os.makedirs(final_dir)

        keyed_element = {element_key: element}  # The element is expected behind its key.

        serialized = serialize_json(keyed_element, json_configurations=json_configurations)

        with open(os.path.join(final_dir, f"{element_key}.json"), "w", encoding="utf-8") as file:
            file.write(serialized + "\n")

    def _write_multi_file_solution(
        self,
        dir_path: str,
        output: Output,
    ):
        """
        Auxiliary function to write the solution files to the specified
        directory.
        """

        if output.solution_files is None:
            return

        solutions_dir = os.path.join(dir_path, SOLUTIONS_KEY)

        if not os.path.exists(solutions_dir):
            os.makedirs(solutions_dir)

        for solution_file in output.solution_files:
            if not isinstance(solution_file, SolutionFile):
                raise TypeError(
                    f"unsupported solution_file type: {type(solution_file)}, supported type is `SolutionFile`"
                )

            file_path = os.path.join(solutions_dir, solution_file.name)
            if solution_file.writer_args is None:
                solution_file.writer_args = []
            if solution_file.writer_kwargs is None:
                solution_file.writer_kwargs = {}

            # Call the writer function with the final path, and user provided
            # arguments and keyword arguments.
            solution_file.writer(
                file_path,
                solution_file.data,
                *solution_file.writer_args,
                **solution_file.writer_kwargs,
            )

    # Callback functions for writing the output data.
    FILE_WRITERS = {
        OutputFormat.JSON: _write_json,
        OutputFormat.CSV_ARCHIVE: _write_archive,
        OutputFormat.MULTI_FILE: _write_multi_file,
        OutputFormat.TEXT: _write_json,
    }
    """Dictionary mapping output formats to writer functions."""

    def write(
        self,
        output: Output | dict[str, Any] | BaseModel,
        path: str | None = None,
        skip_stdout_reset: bool = False,
    ) -> None:
        """
        Write the output to the local filesystem or stdout.

        This method writes the provided output to the specified path or to stdout,
        depending on the output format and the path parameter.

        Parameters
        ----------
        output : Union[Output, dict[str, Any], BaseModel]
            Output data to write. Can be an Output object, a dictionary, or a BaseModel.
        path : str, optional
            Path to write the output data to. The interpretation depends on the output format:
            - For OutputFormat.JSON: File path for the JSON output. If None or empty, writes to stdout.
            - For OutputFormat.CSV_ARCHIVE: Directory path for CSV files. If None or empty,
              writes to a directory named "output" in the current working directory.
        skip_stdout_reset : bool, optional
            Skip resetting stdout before writing the output data. Default is False.

        Raises
        ------
        ValueError
            If the Output.output_format is not supported.
        TypeError
            If the output is of an unsupported type.

        Notes
        -----
        This function detects if stdout was redirected and resets it to avoid
        unexpected behavior. If you want to skip this behavior, set the
        skip_stdout_reset parameter to True.

        If the output is a dict or a BaseModel, it will be written as JSON. If
        the output is an Output object, it will be written according to its
        output_format.

        Examples
        --------
        >>> from nextmv.output import LocalOutputWriter, Output
        >>> writer = LocalOutputWriter()
        >>> # Write JSON to a file
        >>> writer.write(Output(solution={"result": 42}), path="result.json")
        >>> # Write JSON to stdout
        >>> writer.write({"simple": "data"})
        """

        # If the user forgot to reset stdout after redirecting it, we need to
        # do it here to avoid unexpected behavior.
        if sys.stdout is not sys.__stdout__ and not skip_stdout_reset:
            reset_stdout()

        if isinstance(output, Output):
            output_format = output.output_format
        elif isinstance(output, dict):
            output_format = OutputFormat.JSON
        elif isinstance(output, BaseModel):
            output_format = OutputFormat.JSON
        else:
            raise TypeError(
                f"unsupported output type: {type(output)}, supported types are `Output`, `dict`, `BaseModel`"
            )

        output_dict = {}
        if isinstance(output, Output):
            output_dict = output.to_dict()
        elif isinstance(output, BaseModel):
            output_dict = output.to_dict()
        elif isinstance(output, dict):
            output_dict = output
        else:
            raise TypeError(
                f"unsupported output type: {type(output)}, supported types are `Output`, `dict`, `BaseModel`"
            )

        self.FILE_WRITERS[output_format](
            self,
            output=output,
            output_dict=output_dict,
            path=path,
        )


def write_local(
    output: Output | dict[str, Any],
    path: str | None = None,
    skip_stdout_reset: bool = False,
) -> None:
    """
    !!! warning
        `write_local` is deprecated, use `write` instead.

    Write the output to the local filesystem or stdout.

    This is a convenience function for instantiating a `LocalOutputWriter` and
    calling its `write` method.

    Parameters
    ----------
    output : Union[Output, dict[str, Any]]
        Output data to write. Can be an Output object or a dictionary.
    path : str, optional
        Path to write the output data to. The interpretation depends on the
        output format:

        - For `OutputFormat.JSON`: File path for the JSON output. If None or
          empty, writes to stdout.
        - For `OutputFormat.CSV_ARCHIVE`: Directory path for CSV files. If None
          or empty, writes to a directory named "output" in the current working
          directory.
    skip_stdout_reset : bool, optional
        Skip resetting stdout before writing the output data. Default is False.

    Raises
    ------
    ValueError
        If the Output.output_format is not supported.
    TypeError
        If the output is of an unsupported type.

    Notes
    -----
    This function detects if stdout was redirected and resets it to avoid
    unexpected behavior. If you want to skip this behavior, set the
    skip_stdout_reset parameter to True.

    Examples
    --------
    >>> from nextmv.output import write_local, Output
    >>> # Write JSON to a file
    >>> write_local(Output(solution={"result": 42}), path="result.json")
    >>> # Write JSON to stdout
    >>> write_local({"simple": "data"})
    """

    deprecated(
        name="write_local",
        reason="`write_local` is deprecated, use `write` instead",
    )

    writer = LocalOutputWriter()
    writer.write(output, path, skip_stdout_reset)


_LOCAL_OUTPUT_WRITER = LocalOutputWriter()
"""Default LocalOutputWriter instance used by the write function."""


def write(
    output: Output | dict[str, Any] | BaseModel,
    path: str | None = None,
    skip_stdout_reset: bool = False,
    writer: OutputWriter | None = _LOCAL_OUTPUT_WRITER,
) -> None:
    """
    Write the output to the specified destination.

    You can import the `write` function directly from `nextmv`:

    ```python
    from nextmv import write
    ```

    This is a convenience function for writing output data using a provided writer.
    By default, it uses the `LocalOutputWriter` to write to files or stdout.

    Parameters
    ----------
    output : Union[Output, dict[str, Any], BaseModel]
        Output data to write. Can be an Output object, a dictionary, or a BaseModel.
    path : str, optional
        Path to write the output data to. The interpretation depends on the
        output format:

        - For `OutputFormat.JSON`: File path for the JSON output. If None or
          empty, writes to stdout.
        - For `OutputFormat.CSV_ARCHIVE`: Directory path for CSV files. If None
          or empty, writes to a directory named "output" in the current working
          directory.
    skip_stdout_reset : bool, optional
        Skip resetting stdout before writing the output data. Default is False.
    writer : OutputWriter, optional
        The writer to use for writing the output. Default is a
        `LocalOutputWriter` instance.

    Raises
    ------
    ValueError
        If the Output.output_format is not supported.
    TypeError
        If the output is of an unsupported type.

    Examples
    --------
    >>> from nextmv.output import write, Output, OutputFormat
    >>> # Write JSON to a file
    >>> write(Output(solution={"result": 42}), path="result.json")
    >>> # Write CSV archive
    >>> data = {"vehicles": [{"id": 1, "capacity": 100}, {"id": 2, "capacity": 150}]}
    >>> write(Output(output_format=OutputFormat.CSV_ARCHIVE, solution=data), path="output_dir")
    """

    writer.write(output, path, skip_stdout_reset)



"""
This module contains configuration utilities for the Nextmv CLI.
"""

from pathlib import Path
from typing import Any

import yaml

from nextmv.cli.error import error
from nextmv.cloud.client import Client

# Some useful constants.
CONFIG_DIR = Path.home() / ".nextmv"
CONFIG_FILE = CONFIG_DIR / "config.yaml"
API_KEY_KEY = "apikey"
ENDPOINT_KEY = "endpoint"
DEFAULT_ENDPOINT = "api.cloud.nextmv.io"
GO_CLI_PATH = CONFIG_DIR / "nextmv"


def load_config() -> dict[str, Any]:
    """
    Load the current configuration from the config file. Returns an empty
    dictionary if no configuration file exists.

    Returns
    -------
    dict[str, Any]
        The current configuration as a dictionary.
    """

    if not CONFIG_FILE.exists():
        return {}

    with CONFIG_FILE.open() as file:
        config = yaml.safe_load(file)

    if config is None:
        return {}
    return config


def save_config(config: dict[str, Any]) -> None:
    """
    Save the given configuration to the config file.

    Parameters
    ----------
    config : dict[str, Any]
        The configuration to save.
    """

    CONFIG_DIR.mkdir(parents=True, exist_ok=True)

    with CONFIG_FILE.open("w") as file:
        yaml.safe_dump(config, file)


def build_client(profile: str | None = None) -> Client:
    """
    Builds a `cloud.Client` using the API key and endpoint for the given
    profile. If no profile is given, the default profile is used. If either the
    API key or endpoint is missing, an exception is raised. If the config is
    not available, an exception is raised.

    Parameters
    ----------
    profile : str | None
        The profile name to use. If None, the default profile is used.

    Returns
    -------
    Client
        A client configured with the API key and endpoint for the selected
        profile or the default configuration.

    Raises
    ------
    typer.Exit
        If no configuration is found, if the requested profile does not exist,
        or if the API key or endpoint (for either the selected profile or the
        default configuration) is not set or is empty.
    """

    config = load_config()
    if config == {}:
        error("No configuration found. Please run [code]nextmv configuration create[/code].")

    if profile is not None:
        if profile not in config:
            error(f"Profile [bold magenta]{profile}[/bold magenta] does not exist.")

        api_key = config[profile].get(API_KEY_KEY)
        if api_key is None or api_key == "":
            error(f"API key for profile [bold magenta]{profile}[/bold magenta] is not set or is empty.")

        endpoint = config[profile].get(ENDPOINT_KEY)
        if endpoint is None or endpoint == "":
            error(f"Endpoint for profile [bold magenta]{profile}[/bold magenta] is not set or is empty.")
    else:
        api_key = config.get(API_KEY_KEY)
        if api_key is None or api_key == "":
            error("Default API key is not set or is empty.")

        endpoint = config.get(ENDPOINT_KEY)
        if endpoint is None or endpoint == "":
            error("Default endpoint is not set or is empty.")

    return Client(api_key=api_key, url=f"https://{endpoint}")


def obscure_api_key(api_key: str) -> str:
    """
    Obscure an API key for display purposes.

    Parameters
    ----------
    api_key : str
        The API key to obscure.

    Returns
    -------
    str
        The obscured API key.
    """

    if len(api_key) <= 4:
        return "*" * len(api_key)

    return api_key[:2] + "*" * 4 + api_key[-2:]



"""
Provides status enums for Nextmv application runs.

This module defines enumerations for representing the status of a run in a
Nextmv application. It includes a deprecated `Status` enum and the current
`StatusV2` enum.

Classes
-------
Status
    Deprecated status of a run.
StatusV2
    Represents the status of a run.
"""

from enum import Enum


class Status(str, Enum):
    """
    !!! warning
        `Status` is deprecated, use `StatusV2` instead.

    Status of a run.

    You can import the `Status` class directly from `nextmv`:

    ```python
    from nextmv import Status
    ```

    This enum represents the possible states of a run. It is deprecated and
    `StatusV2` should be used for new implementations.

    Attributes
    ----------
    failed : str
        Run failed.
    running : str
        Run is running.
    succeeded : str
        Run succeeded.

    Examples
    --------
    >>> from nextmv.cloud import Status
    >>> current_status = Status.running
    >>> if current_status == Status.succeeded:
    ...     print("Run completed successfully.")
    ... elif current_status == Status.failed:
    ...     print("Run failed.")
    ... else:
    ...     print(f"Run is currently {current_status.value}.")
    Run is currently running.

    """

    failed = "failed"
    """Run failed."""
    running = "running"
    """Run is running."""
    succeeded = "succeeded"
    """Run succeeded."""


class StatusV2(str, Enum):
    """
    Status of a run.

    You can import the `StatusV2` class directly from `nextmv`:

    ```python
    from nextmv import StatusV2
    ```

    This enum represents the comprehensive set of possible states for a run
    in Nextmv.

    Attributes
    ----------
    canceled : str
        Run was canceled.
    failed : str
        Run failed.
    none : str
        Run has no status.
    queued : str
        Run is queued.
    running : str
        Run is running.
    succeeded : str
        Run succeeded.

    Examples
    --------
    >>> from nextmv.cloud import StatusV2
    >>> run_status = StatusV2.queued
    >>> print(f"The run status is: {run_status.value}")
    The run status is: queued

    >>> if run_status == StatusV2.succeeded:
    ...     print("Processing complete.")
    ... elif run_status in [StatusV2.queued, StatusV2.running]:
    ...     print("Processing in progress.")
    ... else:
    ...     print("Processing has not started or has ended with issues.")
    Processing in progress.

    """

    canceled = "canceled"
    """Run was canceled."""
    failed = "failed"
    """Run failed."""
    none = "none"
    """Run has no status."""
    queued = "queued"
    """Run is queued."""
    running = "running"
    """Run is running."""
    succeeded = "succeeded"
    """Run succeeded."""
